<?php

namespace Drupal\opigno_certificate_pdf_form\Controller;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\StreamWrapper\StreamWrapperInterface;
use Drupal\Core\Url;
use Drupal\entity_print\Plugin\EntityPrintPluginManagerInterface;
use Drupal\entity_print\Plugin\ExportTypeManagerInterface;
use Drupal\entity_print\PrintBuilderInterface;
use Drupal\fillpdf\FillPdfFormInterface;
use Drupal\fillpdf\TokenResolverInterface;
use Drupal\fillpdf\Service\BackendProxyInterface;
use Drupal\fillpdf\Plugin\FillPdfActionPluginManager;
use Drupal\group\Entity\GroupInterface;
use Drupal\opigno_certificate\Controller\EntityPrintController as WrappedEntityPrintController;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Shim to send some certificates to Entity Print, and some to fillpdf.
 */
class EntityPrintController extends WrappedEntityPrintController {

  /**
   * The backend proxy.
   *
   * @var \Drupal\fillpdf\Service\BackendProxyInterface
   */
  protected $backendProxy;

  /**
   * The FillPDF token resolver.
   *
   * @var \Drupal\fillpdf\TokenResolverInterface
   */
  protected $tokenResolver;

  /**
   * The FillPDF action manager.
   *
   * @var \Drupal\fillpdf\Plugin\FillPdfActionPluginManager
   */
  protected $actionManager;

  protected LoggerInterface $logger;


  /**
   * {@inheritdoc}
   */
  public function __construct(EntityPrintPluginManagerInterface $plugin_manager, ExportTypeManagerInterface $export_type_manager, PrintBuilderInterface $print_builder, EntityTypeManagerInterface $entity_type_manager, BackendProxyInterface $backend_proxy, TokenResolverInterface $token_resolver, FillPdfActionPluginManager $action_manager, LoggerInterface $logger) {
    $this->backendProxy = $backend_proxy;
    $this->tokenResolver = $token_resolver;
    $this->actionManager = $action_manager;
    $this->logger = $logger;

    parent::__construct($plugin_manager, $export_type_manager, $print_builder, $entity_type_manager);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('plugin.manager.entity_print.print_engine'),
      $container->get('plugin.manager.entity_print.export_type'),
      $container->get('entity_print.print_builder'),
      $container->get('entity_type.manager'),
      $container->get('fillpdf.backend_proxy'),
      $container->get('fillpdf.token_resolver'),
      $container->get('plugin.manager.fillpdf_action.processor'),
      $container->get('logger.factory')->get('aft_membership')
    );
  }

  /**
   * Print an entity to the selected format.
   *
   * @param string $entity_type
   *   The entity type.
   * @param string $entity_id
   *   The entity ID.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   The response object on error otherwise the Print is sent.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function viewPrint($entity_type, $entity_id) {
    // Override the behaviour only for the pdf_form bundle.
    $entity = $this->entityTypeManager->getStorage($entity_type)->load($entity_id);
    $opigno_certificate = $entity->field_certificate->entity;

    if ($opigno_certificate->bundle() !== 'pdf_form') {
      return parent::viewPrint($entity_type, $entity_id);
    }
    else {
      return $this->populatePdf($entity);
    }
  }

  protected function populatePdf($learning_path) {
    $opigno_certificate = $learning_path->field_certificate->entity;
    $fillpdf_form = $opigno_certificate->pdf_form->entity;
    if ($fillpdf_form === NULL) {
      $this->messenger()->addError($opigno_certificate->getEntityTypeId());
      return $this->handleError($learning_path, sprintf('Fillpdf form missing for LMS certificate (go to /certificate/%d/edit).', $opigno_certificate->id()));
    }
    $user = \Drupal::currentUser();

    $lp_status_storage = \Drupal::entityTypeManager()->getStorage('user_lp_status');
    $query = $lp_status_storage->getQuery();
    $results = $query
      ->condition('gid', $learning_path->id())
      ->condition('uid', $user->id())
      ->accessCheck(FALSE)
      ->execute();
    $lp_status = !empty($results) ? $lp_status_storage->load(key($results)) : FALSE;

    // @todo Use TrainingManager and LPStatus to find when the user completed
    // the learning path, or at least make sure a token is available for it.
    $entities = [
      'user' => [$user],
      'group' => [$learning_path],
      'learning_path_status' => [$lp_status],
    ];

    $context = ['entities' => $entities];

    $populated_pdf = $this->backendProxy->merge($fillpdf_form, $entities, $context);
    if (empty($populated_pdf)) {
      return $this->handleError($learning_path);
    }

    $filename = $this->buildFilename($fillpdf_form->title->value, $entities);

    // Fake the $context variable which is usually taken from decoding the
    // request, we only provide the 'entities' key for now which is redundant.
    return $this->handlePopulatedPdf($fillpdf_form, $populated_pdf, ['entities' => $entities], $filename, $context);
  }

  /**
   * Error handler method.
   */
  private function handleError(GroupInterface $learning_path, ?string $admin_message = NULL): RedirectResponse {
    if ($admin_message !== NULL) {
        $this->logger->error($admin_message);
    }
    $this->messenger()->addError($this->t('Unable to generate your certificate, please try again later, or contact an administrator if the problem persists.'));
    return new RedirectResponse(Url::fromRoute('entity.group.canonical', ['group' => $learning_path->id()])->toString());
  }

  /**
   * Builds the filename of a populated PDF file.
   *
   * Copypasta from fillpdf.
   *
   * @param string $original
   *   The original filename without tokens being replaced.
   * @param \Drupal\Core\Entity\EntityInterface[] $entities
   *   An array of entities to be used for replacing tokens.
   *
   * @return string
   *   The token-replaced filename.
   */
  protected function buildFilename($original, array $entities) {
    // Replace tokens *before* sanitization.
    $original = (string) $this->tokenResolver->replace($original, $entities, ['content' => 'text']);

    $output_name = str_replace(' ', '_', $original);
    $output_name = preg_replace('/\.pdf$/i', '', $output_name);
    $output_name = preg_replace('/[^a-zA-Z0-9_.-]+/', '', $output_name) . '.pdf';

    return $output_name;
  }

  /**
   * Figure out what to do with the PDF and do it.
   *
   * Copypasta from fillpdf.
   *
   * @param \Drupal\fillpdf\FillPdfFormInterface $fillpdf_form
   *   An object containing the loaded record from {fillpdf_forms}.
   * @param string $pdf_data
   *   A string containing the content of the merged PDF.
   * @param array $context
   *   The request context as returned by
   *   FillPdfLinkManipulatorInterface::parseLink().
   * @param string $filename
   *   Filename of the merged PDF.
   * @param \Drupal\Core\Entity\EntityInterface[] $entities
   *   An array of entities to be used for replacing tokens. These may be still
   *   needed for generating the destination path, if the file is saved.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   The action plugin's response object.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @see \Drupal\fillpdf\FillPdfLinkManipulatorInterface::parseLink()
   */
  protected function handlePopulatedPdf(FillPdfFormInterface $fillpdf_form, $pdf_data, array $context, $filename, array $entities) {
    $force_download = FALSE;
    if (!empty($context['force_download'])) {
      $force_download = TRUE;
    }

    // Determine the appropriate action for the PDF.
    $scheme = $fillpdf_form->getStorageScheme();
    $is_available = array_key_exists($scheme, \Drupal::service('stream_wrapper_manager')->getWrappers(StreamWrapperInterface::WRITE_VISIBLE));
    $is_allowed = in_array($scheme, \Drupal::config('fillpdf.settings')->get('allowed_schemes') ?: []);

    if (empty($scheme)) {
      $action_plugin_id = 'download';
    }
    elseif (!$is_available || !$is_allowed) {
      // @todo: We don't need the ID once an admin_title is #required,
      // see https://www.drupal.org/project/fillpdf/issues/3040776.
      $label = $fillpdf_form->label() . " ({$fillpdf_form->id()})";
      $this->getLogger('fillpdf')->critical('Saving a generated PDF file in unavailable storage scheme %scheme failed.', [
        '%scheme' => "$scheme://",
      ]);
      if ($this->currentUser()->hasPermission('administer pdfs')) {
        $this->messenger()->addError($this->t('File storage scheme %scheme:// is unavailable, so a PDF file generated from FillPDF form @link could not be stored.', [
          '%scheme' => $scheme,
          '@link' => Link::fromTextAndUrl($label, $fillpdf_form->toUrl())->toString(),
        ]));
      }
      // Make sure the file is only sent to the browser.
      $action_plugin_id = 'download';
    }
    else {
      $redirect = !empty($fillpdf_form->destination_redirect->value);
      $action_plugin_id = $redirect ? 'redirect' : 'save';
    }

    // @todo: Remove in FillPDF 5.x. The filename is not part of the context and
    // is separately available anyway.
    $context['filename'] = $filename;

    // @todo: Rename 'token_objects' to 'entities' in FillPDF 5.x. Webform
    // submissions are now entities, too.
    $action_configuration = [
      'form' => $fillpdf_form,
      'context' => $context,
      'token_objects' => $entities,
      'data' => $pdf_data,
      'filename' => $filename,
    ];

    /** @var \Drupal\fillpdf\Plugin\FillPdfActionPluginInterface $fillpdf_action */
    $fillpdf_action = $this->actionManager->createInstance($action_plugin_id, $action_configuration);
    $response = $fillpdf_action->execute();

    // If we are forcing a download, then manually get a Response from
    // the download action and return that. Side effects of other plugins will
    // still happen, obviously.
    if ($force_download) {
      /** @var FillPdfDownloadAction $download_action */
      $download_action = $this->actionManager
        ->createInstance('download', $action_configuration);
      $response = $download_action
        ->execute();
    }

    return $response;
  }

}
