<?php

declare(strict_types=1);

namespace Drupal\opigno_learning_path\Controller;

use Drupal\Core\Url;
use Drupal\group\Entity\GroupInterface;
use Drupal\opigno_learning_path\Exception\TrainingException;

/**
 * Common methods for training navigation logic classes.
 */
trait TrainingControllerTrait {

  /**
   * Error handler.
   */
  private function handleError(GroupInterface $group, \Exception $e): Url {
    // Training exception - reset current module and activity and redirect
    // to the module.
    if ($e instanceof TrainingException) {
      $iterations = 0;
      // Loop could be used in case of free navigation, otherwise there'll
      // always be a single iteration without additional exceptions.
      do {
        $success = TRUE;
        $lp_status = $e->getLpStatus();
        $module = $e->getModule();
        try {
          if ($e->getCode() === TrainingException::REQUIRED_NOT_FINISHED) {
            $module_status = $this->trainingManager->loadModuleStatus($lp_status->id(), $module->id());
            $message = $this->t('You need to finish module "@module" first.', [
              '@module' => $module->label(),
            ]);
          }
          elseif ($e->getCode() === TrainingException::REQUIRED_NOT_STARTED) {
            $module_status = $this->trainingManager->initializeModule($lp_status, $module);
            $message = $this->t('You need to finish module "@module" first.', [
              '@module' => $module->label(),
            ]);
          }
          // In case of insufficient score the entire module needs to be
          // restarted.
          elseif ($e->getCode() === TrainingException::INSUFFICIENT_SCORE) {
            $module_status = $this->trainingManager->initializeModule($lp_status, $module);
            $message = $this->t('Insufficient score on "@module" to proceed. Please retake the module.', [
              '@module' => $module->label(),
            ]);
          }

          $module_status->save();
          $lp_status->set('current_module_status', $module_status);
          $lp_status->save();
          $this->messenger()->addError($message);
          return Url::fromRoute('opigno_module.group.answer_form', [
            'group' => $group->id(),
            'opigno_module' => $module_status->get('module')->target_id,
            'opigno_activity' => $module_status->getCurrentActivityId(),
          ]);
        }
        catch (TrainingException $nested_exception) {
          $success = FALSE;
          $e = $nested_exception;
        }

        $iterations++;
        if ($iterations > 50) {
          throw new \Exception('Too many iterations, please inspect the code.');
        }
        // First module always can be initialized so no risk of infinitely
        // throwing a TrainingException.
      } while (!$success);
    }

    // Other exception type - redirect to the group, log exception and
    // display a generic message.
    $this->messenger()->addError($this->t('An error occurred while trying to start training. Please report the issue to your site administrator.'));
    \watchdog_exception('opigno_learning_path', $e, 'Training @group_id start error for user @user_id. Message: @message Backtrace: @backtrace_string', [
      '@group_id' => $group->id(),
      '@user_id' => $this->currentUser()->id(),
      '@message' => $e->getMessage(),
    ]);
    return Url::fromRoute('entity.group.canonical', [
      'group' => $group->id(),
    ]);
  }

}
