<?php

declare(strict_types=1);

namespace Drupal\opigno_learning_path;

use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Link;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Render\Markup;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\group\Entity\GroupInterface;
use Drupal\opigno_learning_path\Entity\Bundle\LearningPath;
use Drupal\opigno_learning_path\Entity\LPStatus;
use Symfony\Component\Routing\Exception\RouteNotFoundException;

final class CourseProgressChecker {

  use MessengerTrait;
  use StringTranslationTrait;

  public function __construct(
    private readonly EntityTypeManagerInterface $entityTypeManager,
    private readonly Connection $database,
  ) {}

  public function checkActivityhasProgress(string $activity_id, bool $display_warning = FALSE): bool {
    $module_ids = $this->database->select('opigno_module_relationship', 'omr')
      ->fields('omr', ['parent_id'])
      ->condition('child_id', $activity_id)
      ->execute()
      ->fetchCol();

    if (\count($module_ids) === 0) {
      return FALSE;
    }

    $course_ids = [];
    foreach ($module_ids as $module_id) {
      foreach ($this->checkModuleHasProgress($module_id) as $course_id) {
        $course_ids[$course_id] = $course_id;
      }
    }

    if (\count($course_ids) === 0) {
      return FALSE;
    }

    if ($display_warning) {
      $links = [];
      foreach ($this->entityTypeManager->getStorage('group')->loadMultiple($course_ids) as $course) {
        $links[] = $course->toLink()->toString();
      }
      $this->messenger()->addWarning($this->t('Warning! This activity is a part of learning paths that have already been started by certain students. Deleting it may cause data integrity errors for students who have not completed the following:<br />@links', [
        '@links' => Markup::create(\implode('<br />', $links)),
      ]));
    }

    return TRUE;
  }

  public function checkModuleHasProgress(string $module_id, bool $display_warning = FALSE): array {
    $module_status_query = $this->entityTypeManager->getStorage('user_module_status')->getQuery();
    $status_ids = $module_status_query
      ->accessCheck(FALSE)
      ->condition('module', $module_id)
      ->condition($module_status_query->orConditionGroup()
        ->condition('finished', '0')
        ->condition('evaluated', FALSE)
      )
      ->execute();

    if (\count($status_ids) === 0) {
      return [];
    }

    $module_statuses = $this->entityTypeManager->getStorage('user_module_status')->loadMultiple($status_ids);
    $course_ids = [];
    foreach ($module_statuses as $module_status) {
      $course_id = $module_status->get('lp_status')->entity->get('gid')->target_id;
      $course_ids[$course_id] = $course_id;
    }

    if ($display_warning) {
      foreach ($this->entityTypeManager->getStorage('group')->loadMultiple($course_ids) as $course) {
        $links[] = $course->toLink()->toString();
      }
      $this->messenger()->addWarning($this->t('Warning! This module is a part of learning paths that have already been started by certain students. Deleting it may cause data integrity errors for students who have not completed the following:<br />@links', [
        '@links' => Markup::create(\implode('<br />', $links)),
      ]));
    }

    return $course_ids;
  }

  public function checkCourseProgress(GroupInterface $course, bool $display_warning = FALSE): bool {
    if (!$course instanceof LearningPath) {
      return FALSE;
    }

    $query = $this->entityTypeManager->getStorage('user_lp_status')->getQuery();
    $in_progress_count = $query
      ->accessCheck(FALSE)
      ->condition('gid', $course->id())
      ->condition('status', [
        LPStatus::STATUS_PROGRESS,
        LPStatus::STATUS_NEEDS_EVALUATION,
      ], 'IN')
      ->count()
      ->execute();
    if ($in_progress_count !== 0) {
      if ($display_warning) {
        $url = Url::fromRoute('view.students.page', [
          'group' => $course->id(),
        ]);
        try {
          $link = Link::fromTextAndUrl($this->t('here'), $url)->toString();
        }
        catch (RouteNotFoundException $e) {
          $this->messenger()->addWarning($this->t('Warning! The @training training has @count students in progress or awaiting grading. Making changes may result in data integrity errors for students.', [
            '@training' => Markup::create($course->toLink()->toString()),
            '@count' => $in_progress_count,
          ]));
          return TRUE;
        }

        $this->messenger()->addWarning($this->t('Warning! The @training training has @count students in progress or awaiting grading. Making changes may result in data integrity errors for students.  If you need to make changes, it is advised to reset those students progress @link.', [
          '@training' => Markup::create($course->toLink()->toString()),
          '@count' => $in_progress_count,
          '@link' => Markup::create($link),
        ]));
      }
      return TRUE;
    }
    return FALSE;
  }

}
