<?php

namespace Drupal\opigno_learning_path\Entity\Bundle;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\group\Access\GroupAccessResult;
use Drupal\group\Entity\Group;

/**
 * Learning path group bundle class.
 */
class LearningPath extends Group {

  /**
   * Get learning path classes.
   *
   * @return \Drupal\group\Entity\Group[]
   *   Class groups array.
   */
  public function getClasses(): array {
    $classes = [];
    $contents = $this->getRelationships('subgroup:opigno_class');
    foreach ($contents as $content) {
      $classes[] = $content->getEntity();
    }
    return $classes;
  }

  /**
   * Class membership checker.
   *
   * Checks if the given account is a member of any of the classes of this
   * learning path.
   */
  public function isClassMember(AccountInterface $account, array &$cache_contexts = []): bool {
    $classes = $this->getClasses();

    foreach ($classes as $class) {
      if ($class->getMember($account) !== FALSE) {
        // If the user is a member of one class, only this single class
        // cache context is needed.
        $cache_contexts = ['user.is_group_member:' . $class->id()];
        return TRUE;
      }
      $cache_contexts[] = 'user.is_group_member:' . $class->id();
    }

    return FALSE;
  }

  /**
   * Check if the given user can take this learning path.
   */
  public function takeAccess(AccountInterface $account): AccessResult {

    // Allow anyone with edit access to the group, whether a member or not, to
    // take the learning path. This is also used for view access.
    // Allow access to any member of the learning path that can edit group
    // for testing purposes.
    $access_result = GroupAccessResult::allowedIfHasGroupPermission($this, $account, 'edit group');
    if ($access_result->isAllowed()) {
      return $access_result;
    }

    // Or if the user is a member of any class.
    $cache_contexts = [];
    $is_class_member = $this->isClassMember($account, $cache_contexts);
    if ($is_class_member) {
      return AccessResult::allowed()
        ->addCacheContexts($cache_contexts);
    }

    // We need to add a membership list cache tags otherwise - for
    // LP and classes.
    $tags = [];
    foreach (\Drupal::entityTypeManager()->getStorage('group_content_type')->loadByProperties([
      'group_type' => ['learning_path', 'opigno_class'],
      'content_plugin' => 'group_membership',
    ]) as $membership_bundle) {
      $tags[] = 'group_content_list:' . $membership_bundle->id();
    }

    return AccessResult::forbidden()
      ->addCacheTags($tags);
  }

  /**
   * Check if user needs to create a membership request to join a class.
   */
  public function requiresValidation(): bool {
    if (!$this->hasField('field_requires_validation')) {
      return FALSE;
    }

    return (bool) $this->get('field_requires_validation')->value;
  }

  /**
   * Check if arbitrary activity is allowed on this LP.
   *
   * @param bool $answered
   *   Has the current question already been answered?
   */
  public function allowArbitraryActivity(bool $answered): bool {
    if (!$this->hasGuidedNavigation()) {
      return TRUE;
    }

    if ($this->noFailMode() && $answered) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Check if group has guided navigation enabled.
   */
  public function hasGuidedNavigation(): bool {
    if (!$this->hasField('field_guided_navigation')) {
      return TRUE;
    }

    $gn_field = $this->get('field_guided_navigation');
    if ($gn_field->isEmpty()) {
      return TRUE;
    }

    return (bool) $gn_field->value;
  }

  /**
   * Get LPStatus loading conditions when starting / continuing a training.
   *
   * If no LPStatus meeting the given conditions is found, training will be
   * (re) initialized.
   *
   * @return mixed[]
   *   Conditions for loading status entity when starting a learning path.
   */
  public function getInitialLpStatusConditions(): array {
    $conditions = [];
    // Always allow to continue if no fail mode is enabled.
    if (!$this->noFailMode()) {
      $conditions['finished'] = 0;
    }
    return $conditions;
  }

  /**
   * Check wether the course has no fail mode enabled.
   */
  public function noFailMode(): bool {
    return (bool) $this->get('no_fail_mode')->value;
  }

  /**
   * {@inheritdoc}
   */
  public static function bundleFieldDefinitions(EntityTypeInterface $entity_type, $bundle, array $base_field_definitions): array {
    $fields = parent::bundleFieldDefinitions($entity_type, $bundle, $base_field_definitions);
    $fields['no_fail_mode'] = BaseFieldDefinition::create('boolean')
      ->setLabel(new TranslatableMarkup('Allow to revisit finished training'))
      ->setDescription(new TranslatableMarkup('Always allow to go back to the training, navigate freely and correct answers without the need to restart.'))
      ->setRevisionable(TRUE)
      ->setTranslatable(FALSE)
      ->setDefaultValue(TRUE)
      ->setDisplayOptions('form', [
        'type' => 'boolean_checkbox',
        'settings' => [
          'display_label' => TRUE,
        ],
      ])
      ->setDisplayConfigurable('form', TRUE);

    return $fields;
  }

}
