<?php

namespace Drupal\opigno_learning_path;

use Drupal\Core\Database\Database;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\group\Entity\Group;
use Drupal\group\Entity\GroupInterface;
use Drupal\opigno_group_manager\Entity\OpignoGroupManagedContent;
use Drupal\opigno_group_manager\OpignoGroupContext;
use Drupal\opigno_module\Entity\OpignoModule;
use Drupal\user\Entity\User;
use Drupal\user\UserInterface;

/**
 * Class LearningPathAccess.
 *
 * @package Drupal\opigno_learning_path
 */
class LearningPathAccess {

  /**
   * Returns group or class subgroup membership.
   *
   * @param \Drupal\group\Entity\GroupInterface $group
   *   The group.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The user account.
   *
   * @return
   *   \Drupal\group\GroupMembership|bool
   */
  public static function getGroupOrClassMembership(GroupInterface $group, AccountInterface $account) {
    $membership = $group->getMember($account);
    if ($membership) {
      return $membership;
    }
    else {
      return static::getClassMembership($group, $account);
    }
  }

  /**
   * Returns class subgroup membership.
   *
   * @param \Drupal\group\Entity\GroupInterface $group
   *   The group.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The user account.
   *
   * @return
   *   Drupal\group\GroupMembership|bool
   */
  public static function getClassMembership(GroupInterface $group, AccountInterface $account) {
    $classes = $group->getRelationships('subgroup:opigno_class');
    foreach ($classes as $class) {
      $membership = $class->getEntity()->getMember($account);
      if ($membership) {
        return $membership;
      }
    }
    return FALSE;
  }

  /**
   * Gets group role.
   */
  public static function getGroupRoles(Group $group) {
    $type = array_shift($group->type->getValue());
    $properties = [
      'group_type' => $type['target_id'],
      'permissions_ui' => TRUE,
    ];
    return \Drupal::entityTypeManager()
      ->getStorage('group_role')
      ->loadByProperties($properties);
  }

  /**
   * Set visibility fields on Learning Path group save.
   */
  public static function setVisibilityFields(Group &$group) {
    if ($group->field_learning_path_visibility->value === 'private') {
      $group->set('field_anonymous_visibility', 0);
      $group->set('field_requires_validation', 1);
    }
  }

  /**
   * Returns LMS course/module access flag.
   */
  public static function getGroupContentAccess(EntityInterface $entity, AccountInterface $account) {
    $access = TRUE;

    $contentID = OpignoGroupContext::getCurrentGroupContentId();
    if (!empty($contentID)) {
      $contents = OpignoGroupManagedContent::loadByProperties(['id' => $contentID]);

      if ($contents = array_shift($contents)) {
        $learningPath = $contents->getGroup();
        if (isset($learningPath)
          && $learningPath->getEntityTypeId() === 'learning_path'
          && !$learningPath->access('view', $account)) {
          $access = FALSE;
        }
      }
    }

    return $access;
  }

  /**
   * Returns group required validation.
   */
  public static function requiredValidation(Group $group, $account = FALSE) {
    if (!$account) {
      $account = \Drupal::currentUser();
    }
    // Check if we need to wait validation.
    $validation = FALSE;
    if ($group->get('field_requires_validation')->value && !$account->hasPermission('bypass group validation requirement')) {
      $validation = TRUE;
    }

    return $validation;
  }

  /**
   * Returns group user access flag in validation condition.
   */
  public static function statusGroupValidation(Group $group, AccountInterface $account) {
    $access = TRUE;
    $status = FALSE;
    if (!$account instanceof UserInterface) {
      $user = User::load($account->id());
    }
    $status = LearningPathAccess::getMembershipStatus($group, $user);

    if ($status) {
      $visibility = $group->field_learning_path_visibility->value;

      // Check if we need to wait validation.
      $validation = LearningPathAccess::requiredValidation($group, $account);

      // @todo: for performance move this to validation on group
      // membership/requests.
      // https://www.drupal.org/project/opigno_learning_path/issues/3275804
      $required_trainings = self::hasUncompletedRequiredTrainings($group, $account);

      if ($visibility === 'public' && $validation) {
        // For public groups with validation status should be 'Active'.
        if ($status != 1) {
          $access = FALSE;
        }
      }
      // If user unfinished required trainings.
      elseif ($required_trainings) {
        $access = FALSE;
      }
      else {
        // For another groups status should be not 'Blocked'.
        if ($status == 3) {
          $access = FALSE;
        }
      }

      // Added possibility to override access for groups from other modules.
      if (!self::triggerHookAccess('status_group_validation_access', $group, $account)) {
        $access = FALSE;
      }

    }

    return $access;
  }

  /**
   * Returns step user access.
   */
  public static function checkStepValidation($step, $uid) {

    $account = User::load($uid);
    switch ($step['typology']) {
      case 'Course':
        $group = Group::load($step['id']);
        return self::triggerHookAccess('status_course_validation_access', $group, $account);
        break;
      case 'Module':
        $module = OpignoModule::load($step['id']);
        return self::triggerHookAccess('status_module_validation_access', $module, $account);
        break;
    }

    return TRUE;
  }

  /**
   * Trigger hook accesses.
   */
  public static function triggerHookAccess($hook_name, $entity, $account) {
    $access = TRUE;
    $access_results = \Drupal::moduleHandler()->invokeAll($hook_name, [$entity, $account]);
    if (is_array($access_results)) {
      $access_results = array_filter($access_results, function($var){
          return empty($var);
      });
      if (!empty($access_results)) {
        $access = FALSE;
      }
    }
    return $access;
  }

  /**
   * Returns user membership statuses array.
   */
  public static function getMembershipStatusesArray() {
    return [
      '1' => t('Active'),
      '2' => t('Pending'),
      '3' => t('Blocked'),
    ];
  }

  /**
   * Get the group membership request status for a user.
   *
   * @param Drupal\group\Entity\GroupInterface
   *   A group entity.
   * @param Drupal\user\Entity\UserInterface;
   *   A user entity.
   *
   * @return string|null
   *   The membership request status, or NULL.
   */
  public static function getMembershipRequestStatus(GroupInterface $group, UserInterface $user): ?string {
    $membership_request_manager = \Drupal::service('grequest.membership_request_manager');
    $membership_request = $membership_request_manager->getMembershipRequest($user, $group);
    if ($membership_request) {
      return $membership_request->grequest_status->value;
    }
    return NULL;
  }

  /**
   * Returns user membership status.
   *
   * @param GroupInterface $group
   *   The group object. Also accepts a group_content membership ID
   *   (deprecated).
   * @param UserInterface $user
   *   The user object.
   * @param bool $as_string
   *   Whether the status should be returned as an interface string.
   */
  public static function getMembershipStatus($group, $user = NULL, $as_string = FALSE) {
    if (!$group instanceof GroupInterface) {
      @trigger_error('Passing a membership group content ID to LearningPathAccess::getMembershipStatus() is deprecated in opigno_learning_path:3.1.0 and will be removed in opigno_learning_path:4.0.0, use ::getMembershipStatus($group, $user, $as_string) instead', E_USER_DEPRECATED);
      $mid = $group;
      $as_string = $user;
      $result = 1;
    }
    else {
      $result = FALSE;
      $membership = static::getGroupOrClassMembership($group, $user);
      if ($membership) {
        $result = 1;
      }
      else {
        if (!$user instanceof UserInterface) {
          $user = User::load($user->id());
        }
        $value = static::getMembershipRequestStatus($group, $user);
        if ($value === 'rejected') {
          $result = 3;
        }
        elseif ($value === 'pending') {
          $result = 2;
        }
      }
    }

    if ($as_string && $result != FALSE) {
      $statuses = LearningPathAccess::getMembershipStatusesArray();
      $result = $statuses[$result];
    }

    return $result;
  }

  /**
   * Returns user subscription events.
   */
  public static function getUserMemberEvents() {
    return [
      '1' => 'subscribed',
      '2' => 'approval',
      '3' => 'blocked',
    ];
  }

  /**
   * Creates database table for Learning Path group user statuses.
   */
  public static function createUserStatusTable() {
    $schema = Database::getConnection()->schema();
    if (!$schema->tableExists('opigno_learning_path_group_user_status')) {
      $spec = [
        'description' => 'Learning Path group user statuses',
        'fields' => [
          'id' => [
            'type' => 'serial',
            'not null' => TRUE,
          ],
          'mid' => [
            'description' => 'Membership ID',
            'type' => 'int',
            'unsigned' => TRUE,
            'not null' => TRUE,
          ],
          'uid' => [
            'description' => 'User ID',
            'type' => 'int',
            'unsigned' => TRUE,
            'not null' => TRUE,
          ],
          'gid' => [
            'description' => 'Group ID',
            'type' => 'int',
            'unsigned' => TRUE,
            'not null' => TRUE,
          ],
          'status' => [
            'description' => 'Member status',
            'type' => 'int',
            'unsigned' => TRUE,
            'not null' => TRUE,
            'default' => 0,
          ],
          'message' => [
            'description' => 'User message',
            'type' => 'varchar',
            'length' => 200,
            'not null' => TRUE,
            'default' => '',
          ],
        ],
        'primary key' => ['id'],
        'indexes' => [
          'mid' => ['mid'],
          'gid' => ['gid'],
          'uid' => ['uid'],
        ],
        'foreign keys' => [
          'group_relationship' => ['mid' => 'id'],
          'users' => ['uid' => 'uid'],
          'groups' => ['gid' => 'id'],
        ],
      ];
      $schema->createTable('opigno_learning_path_group_user_status', $spec);
    }
  }

  /**
   * Returns uncompleted required trainings flag.
   *
   * @param \Drupal\group\Entity\GroupInterface $group
   *   Group.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   User account.
   *
   * @return bool|array
   *   Return FALSE or Array with unfinished training.
   */
  public static function hasUncompletedRequiredTrainings(GroupInterface $group, AccountInterface $account) {
    $trainings = $group->get('field_required_trainings')->getValue();
    if (!$trainings) {
      return FALSE;
    }

    $database = \Drupal::database();
    // Array of uncompleted required trainings.
    $uncompleted = [];
    if (is_array($trainings)) {
      for ($i = 0; $i < count($trainings); $i++) {
        $gid = $trainings[$i]['target_id'];
        // Check that training is completed.
        $is_completed = $database->select('opigno_learning_path_achievements', 'a')
          ->fields('a')
          ->condition('uid', $account->id())
          ->condition('gid', $gid)
          ->condition('status', 'completed')
          ->execute()->fetchObject();
        if (!$is_completed) {
          array_push($uncompleted, $gid);
        }
      }
    }

    return $uncompleted;
  }

  /**
   * Check if user has needed local role.
   *
   * @param string $role
   *   User role in group.
   *
   * @param \Drupal\Core\Session\AccountInterface $account
   *   User account.
   *
   * @param int $gid
   *   Group ID.
   *
   * @return bool
   *   Check if user has access to Class or Course.
   */
  public static function memberHasRole($role, AccountInterface $account, $gid = NULL) {
    $role = "learning_path-{$role}";

    $connection = Database::getConnection();
    $query = $connection->select('group_relationship_field_data', 'g_c_f_d')
      ->fields('g_c_f_d', ['gid']);
    $query->leftJoin('group_content__group_roles', 'g_c_g_r', 'g_c_f_d.id = g_c_g_r.entity_id');
    $query->condition('g_c_g_r.group_roles_target_id', $role);
    $query->condition('g_c_f_d.entity_id', $account->id());
    $query->condition('g_c_f_d.type', 'learning_path-group_membership');
    if (!empty($gid)) {
      $query->condition('g_c_f_d.gid', $gid);
    }
    $lp_counts = $query->countQuery()->execute()->fetchField();

    return $lp_counts > 0;
  }

  /**
   * Replaces tokens.
   */
  public static function replaceGroupUserTokens(&$text, $params, $token) {
    // @todo: token replace is always available, there's no need to check if the
    // contributed token module is installed.
    // @see https://www.drupal.org/project/opigno_learning_path/issues/3280224
    if ($token) {
      $text = \Drupal::token()->replace($text);
    }

    $text = str_replace([
      '[user]',
      '[group]',
      '[link]',
      '[user-status]',
    ], [
      $params['account']->getAccountName(),
      $params['group']->label(),
      $params['link'],
      $params['status'],
    ], $text);
  }

}
