<?php

declare(strict_types=1);

namespace Drupal\opigno_learning_path\Plugin\Block;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Block\BlockBase;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Drupal\opigno_learning_path\Entity\Bundle\LearningPath;
use Drupal\opigno_learning_path\TrainingManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a 'article' block.
 *
 * @Block(
 *   id = "lp_steps_block",
 *   admin_label = @Translation("LP Steps block"),
 *   context_definitions = {
 *     "user" = @ContextDefinition("entity:user", label = @Translation("User")),
 *     "group" = @ContextDefinition("entity:group", label = @Translation("Group")),
 *     "opigno_module" = @ContextDefinition("entity:opigno_module", label = @Translation("Module")),
 *     "opigno_activity" = @ContextDefinition("entity:opigno_activity", label = @Translation("Activity")),
 *   }
 * )
 */
class StepsBlock extends BlockBase implements ContainerFactoryPluginInterface {

  private const CACHE_PREFIX = 'steps_block_data_cache_';

  /**
   * Static cache of user module statuses.
   */
  private ?array $moduleStatuses = NULL;

  /**
   * Data structure static cache.
   */
  private ?array $structure = NULL;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    protected readonly AccountInterface $currentUser,
    protected readonly CacheBackendInterface $cache,
    protected readonly TrainingManager $trainingManager
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('current_user'),
      $container->get('cache_factory')->get('data'),
      $container->get('opigno_learning_path.training_manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheMaxAge() {
    // This block changes on almost every request except navigation
    // to already completed activities which happens rarely.
    // @todo: implement marking of current module and current activity
    // in frontend, then we'll be able to cache this block and invalidate
    // when a module status is updated / created for this group and user
    // (custom cache tag). Changes to training itself will be covered after
    // we will implement status cleanup upon making such changes.
    return 0;
  }

  private function setModuleStatuses(LearningPath $group, array $module_ids): void {
    if ($this->moduleStatuses !== NULL) {
      return;
    }

    $this->moduleStatuses = [];
    if (!$group instanceof LearningPath) {
      return;
    }

    $lp_status = $this->trainingManager->getLearningPathStatus($group, $this->getContextValue('user'));
    if ($lp_status === NULL) {
      return;
    }

    $module_statuses = $this->trainingManager->loadModuleStatusMultiple($lp_status->id(), $module_ids);
    foreach ($module_statuses as $module_status) {
      $this->moduleStatuses[$module_status->get('module')->target_id] = $module_status;
    }
  }

  /**
   * Get data structure for this training.
   */
  private function getStructure(LearningPath $group): array {
    if ($this->structure !== NULL) {
      return $this->structure;
    }
    $cid = self::CACHE_PREFIX . $group->id();
    $cache = $this->cache->get($cid);
    if ($cache !== FALSE) {
      $this->structure = $cache->data;
    }
    else {
      $this->structure = $this->trainingManager->getOrderedSteps($group);
      $invalidate_tags = [];
      foreach (\array_keys($this->structure) as $module_id) {
        $invalidate_tags[] = 'opigno_module:' . $module_id;
      }
      $this->cache->set($cid, $this->structure, Cache::PERMANENT, $invalidate_tags);
    }

    return $this->structure;
  }

  /**
   * {@inheritdoc}
   */
  protected function blockAccess(AccountInterface $account) {
    $group = $this->getContextValue('group');
    if (!$group instanceof LearningPath) {
      return AccessResult::forbidden();
    }

    $structure = $this->getStructure($group);
    $invalidate_tags = [];
    foreach (\array_keys($structure) as $module_id) {
      $invalidate_tags[] = 'opigno_module:' . $module_id;
    }

    $result = AccessResult::allowed();

    // Don't display the block if there's a single activity only.
    if (\count($structure) === 1) {
      $item = \reset($structure);
      if (\count($item['activities']) === 1) {
        $result = AccessResult::forbidden();
      }
    }
    return $result->addCacheableDependency($group)->addCacheTags($invalidate_tags);
  }

  /**
   * {@inheritdoc}
   */
  public function build() {
    $group = $this->getContextValue('group');
    if (!$group instanceof LearningPath) {
      return [];
    }

    $structure = $this->getStructure($group);

    $this->setModuleStatuses($group, \array_keys($structure));

    $renderable = [
      '#type' => 'container',
      [
        '#theme' => 'opigno_lp_step_activity',
        'title' => [
          '#markup' => $group->label(),
        ],
        'steps' => [],
      ],
    ];

    foreach ($structure as $module_id => $step_data) {
      $module_id = (string) $module_id;
      $step_renderable = [
        '#theme' => 'opigno_lp_step_module',
        '#label' => $step_data['label'],
        '#state' => 'pending',
        '#current' => FALSE,
        '#link' => FALSE,
        '#locked' => \count($step_data['activities']) === 0,
        'activities' => [],
      ];

      $cache_tags = ['opigno_module:' . $module_id];
      foreach (\array_keys($step_data['activities']) as $activity_id) {
        $cache_tags[] = 'opigno_activity:' . $activity_id;
      }

      $this->mergeActivityRenderables($step_renderable['activities'], $step_data['activities'], $group->id(), $module_id);

      $cacheability = new CacheableMetadata();
      $cacheability->addCacheTags($cache_tags);
      $cacheability->applyTo($step_renderable);

      $this->applyModuleStatusData($step_renderable, $step_data, $group->id(), $module_id);

      $renderable[0]['steps'][] = $step_renderable;
    }

    return $renderable;
  }

  /**
   * Merge activity renderable to activities renderable.
   */
  private function mergeActivityRenderables(array &$activities, array $activities_data, string $group_id, string $module_id) {
    foreach ($activities_data as $activity_id => $activity_data) {
      $item = $activities[$activity_id] ?? NULL;
      if ($item === NULL) {
        $item = [
          '#theme' => 'opigno_lp_step_module_activity',
          '#state' => $activity_data['state'] ?? 'pending',
          '#current' => $activity_data['current'] ?? FALSE,
          '#link' => Url::fromRoute('opigno_module.group.answer_form', [
            'group' => $group_id,
            'opigno_activity' => $activity_id,
            'opigno_module' => $module_id,
          ])->toString(),
          '#title' => $activity_data['label'],
          '#activity' => $activity_data['activity'],
        ];
      }
      else {
        $item['#state'] = $activity_data['state'];
        $item['#current'] = $activity_data['current'];
      }
      $activities[$activity_id] = $item;
    }
  }

  /**
   * Apply module status data to module renderable.
   */
  private function applyModuleStatusData(array &$step_renderable, array $step_data, string $group_id, string $module_id): void {
    if (!\array_key_exists($module_id, $this->moduleStatuses)) {
      return;
    }

    $module_status = $this->moduleStatuses[$module_id];

    // If a module status exists, no point caching that at all.
    unset($step_renderable['#cache']);

    if ($module_status->isFinished()) {
      if ($module_status->getScore() >= $step_data['required_score']) {
        $step_renderable['#state'] = 'passed';
      }
      else {
        $step_renderable['#state'] = 'failed';
      }
    }

    $step_renderable['#current'] = $module_id === $this->getContextValue('opigno_module')->id();
    $current_activity = $this->getContextValue('opigno_activity');

    $activities_data = [];
    $activities_field = $module_status->get('activities');
    if (!$activities_field->isEmpty()) {
      $step_renderable['#locked'] = FALSE;
      $activity_ids = [];
      foreach ($activities_field as $item) {
        $activity_ids[$item->target_id] = $item->target_id;
      }
      $answers = $this->trainingManager->getAnswersByActivityId($module_status, $activity_ids);
      foreach ($activities_field as $item) {
        $activity = $item->entity;
        if ($activity === NULL) {
          continue;
        }
        $activities_data[$activity->id()] = [
          'label' => $activity->label(),
          'state' => \array_key_exists($activity->id(), $answers) ? 'passed' : 'pending',
          'current' => $step_renderable['#current'] && ($current_activity->id() === $activity->id()),
          'activity' => $activity,
        ];
      }
    }

    $this->mergeActivityRenderables($step_renderable['activities'], $activities_data, $group_id, $module_id);
  }

}
