<?php

declare(strict_types=1);

namespace Drupal\opigno_learning_path\Plugin\GroupContentCloner;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\IntegrityConstraintViolationException;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\group\Entity\GroupInterface;
use Drupal\group\Entity\GroupType;
use Drupal\group\Plugin\Group\Relation\GroupRelationTypeManagerInterface;
use Drupal\group_clone\Plugin\GroupContentCloner\GroupContentClonerBase;
use Drupal\h5p\H5PDrupal\H5PDrupal;
use Drupal\opigno_module\Entity\OpignoModule;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Filesystem\Filesystem;

/**
 * Default group content cloner plugin.
 *
 * @GroupContentCloner(
 *   id = "opigno_group_managed_content",
 *   label = "Group managed content",
 *   content_plugin_id = "_opigno_group_content",
 * )
 */
class OpignoManagedContentCloner extends GroupContentClonerBase {

  private GroupRelationTypeManagerInterface $groupContentEnablerManager;
  private Connection $database;
  private Filesystem $fileSystem;

  /**
   * Constructs a new instance.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    array $plugin_definition,
    EntityTypeManagerInterface $entity_type_manager,
    GroupRelationTypeManagerInterface $group_content_enabler_manager,
    Connection $database
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager);

    $this->groupContentEnablerManager = $group_content_enabler_manager;
    $this->database = $database;
    $this->fileSystem = new Filesystem();
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): self {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('group_relation_type.manager'),
      $container->get('database')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getConfigurationForm(
    &$element,
    FormStateInterface $form_state,
    GroupType $group_type,
    ?GroupInterface $group = NULL
  ): void {
    if ($group === NULL) {
      return;
    }

    $content_ids = $this->getContentIdsToClone($group);
    if (\count($content_ids['opigno_group_content']) === 0) {
      return;
    }

    $contents = $this->entityTypeManager->getStorage('opigno_group_content')->loadMultiple($content_ids['opigno_group_content']);
    $element['modules'] = [
      '#type' => 'details',
      '#title' => $this->t('Modules'),
      '#open' => TRUE,
      '#tree' => TRUE,
    ];
    foreach ($contents as $content_id => $entity) {
      // @todo Replace this with $entity->module->entity->label() when
      // courses are removed.
      $module = $this->entityTypeManager->getStorage('opigno_module')->load($entity->get('entity_id')->value);
      $element['modules'][$content_id] = [
        '#type' => 'radios',
        '#title' => $module->label(),
        '#options' => [
          'references' => $this->t('Clone references'),
          'entities' => $this->t('Clone entities'),
        ],
        '#default_value' => 'entities',
      ];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getContentIdsToClone(GroupInterface $group): array {
    $clonables = [];

    $query = $this->entityTypeManager->getStorage('opigno_group_content')->getQuery();
    $clonables['opigno_group_content'] = $query
      ->accessCheck(FALSE)
      ->condition('group_id', $group->id())
      ->execute();

    return $clonables;
  }

  /**
   * {@inheritdoc}
   */
  public function appliesTo(GroupType $group_type): bool {
    return $group_type->id() === 'learning_path';
  }

  /**
   * {@inheritdoc}
   */
  public function clone(string $type, string $id, string $group_clone_id, array &$cloned_entity_ids): void {
    $opigno_group_content = $this->entityTypeManager->getStorage('opigno_group_content')->load($id);

    $module_id = $cloned_module_id = $opigno_group_content->get('entity_id')->value;
    if (
      \array_key_exists($id, $this->configuration['modules']) &&
      $this->configuration['modules'][$id] === 'entities'
    ) {
      // @todo Convert to ->entity when courses are removed and the field is
      // converted to proper entity_reference.
      $opigno_module = $this->entityTypeManager->getStorage('opigno_module')->load($module_id);
      $module_clone = $this->cloneModule($opigno_module, $group_clone_id, $cloned_entity_ids);
      $cloned_module_id = $module_clone->id();
    }

    // Clone opigno_module_group plugin content as well
    // (used by group manager only).
    $group_content_storage = $this->entityTypeManager->getStorage('group_content');
    $query = $group_content_storage->getQuery();
    $group_content_ids = $this->groupContentEnablerManager->getRelationshipTypeIds('opigno_module_group');
    $opigno_module_group_ids = $query
      ->accessCheck(FALSE)
      ->condition('gid', $opigno_group_content->get('group_id')->target_id)
      ->condition('entity_id', $module_id)
      ->condition('type', $group_content_ids, 'IN')
      ->execute();
    foreach ($opigno_module_group_ids as $opigno_module_group_id) {
      $opigno_module_group = $group_content_storage->load($opigno_module_group_id);
      $this->cloneEntity($opigno_module_group, $cloned_entity_ids, 'references', [
        'gid' => $group_clone_id,
        'entity_id' => $cloned_module_id,
      ]);
    }

    // Clone standard group content.
    $opigno_group_content_clone = $this->cloneEntity($opigno_group_content, $cloned_entity_ids, 'references', [
      'group_id' => $group_clone_id,
      'entity_id' => $cloned_module_id,
    ]);

    // Clone all links referencing this entity.
    if (!\array_key_exists('opigno_group_link', $cloned_entity_ids)) {
      $cloned_entity_ids['opigno_group_link'] = [];
    }
    $link_storage = $this->entityTypeManager->getStorage('opigno_group_link');
    $original_group_id = $opigno_group_content->get('group_id')->target_id;

    foreach (['parent_content_id', 'child_content_id'] as $link_field) {
      $link_ids = $link_storage
        ->getQuery()
        ->accessCheck(FALSE)
        // We need to load both original and cloned links as there are 2
        // properties to update for each link on different content entities.
        ->condition('group_id', [$original_group_id, $group_clone_id], 'IN')
        ->condition($link_field, $id)
        ->execute();
      $links = $link_storage->loadMultiple($link_ids);

      foreach ($links as $link) {
        // It may be that this link was already cloned throught the child
        // relationship, in such a case only update the parent property and
        // save.
        if (\array_key_exists($link->id(), $cloned_entity_ids['opigno_group_link'])) {
          $cloned_link = $link_storage->load($cloned_entity_ids['opigno_group_link'][$link->id()]);
          $cloned_link->set($link_field, $opigno_group_content_clone->id());
          $cloned_link->save();
        }
        else {
          $this->cloneEntity($link, $cloned_entity_ids, 'references', [
            $link_field => $opigno_group_content_clone->id(),
            'group_id' => $group_clone_id,
          ]);
        }
      }
    }
  }

  /**
   * Clone LMS module.
   */
  public function cloneModule(OpignoModule $original_module, string $group_clone_id, array &$cloned_entity_ids) {
    $cloned_module = $original_module->createDuplicate();
    $cloned_module->get('created')->applyDefaultValue();
    $cloned_module->get('changed')->applyDefaultValue();
    $cloned_module->save();
    $cloned_entity_ids['opigno_module'][$original_module->id()] = $cloned_module->id();

    // Clone activities.
    // @todo Use training manager once ELN-1034 is merged to main.
    $db_activity_data = $original_module->getModuleActivities();
    if (\count($db_activity_data) === 0) {
      return $cloned_module;
    }
    $activity_data = [];
    foreach ($db_activity_data as $item) {
      $activity_data[$item->id] = [
        'max_score' => $item->max_score,
        'group_id' => $group_clone_id,
        'activity_status' => $item->activity_status,
        'auto_update_max_score' => $item->auto_update_max_score,
      ];
    }
    $original_activities = $this->entityTypeManager->getStorage('opigno_activity')->loadMultiple(\array_keys($activity_data));

    $relationship_values = [];
    $weight = -1000;
    $h5p_storage = $this->entityTypeManager->getStorage('h5p_content');
    $h5p = H5PDrupal::getInstance('interface');
    $h5p_path = $h5p->getOption('default_path', 'h5p');
    foreach ($original_activities as $original_activity) {
      $cloned_activity = $original_activity->createDuplicate();
      $cloned_activity->get('created')->applyDefaultValue();
      $cloned_activity->get('changed')->applyDefaultValue();
      // We also need to clone h5p entities, as shared entities are deleted
      // when any of the referencing entities are deleted leading to errors.
      if ($original_activity->hasField('opigno_h5p')) {
        $h5p_field = $original_activity->get('opigno_h5p');
        if (!$h5p_field->isEmpty()) {
          $h5p_entity = $h5p_storage->load($h5p_field->h5p_content_id);
          $h5p_entity_clone = $this->cloneEntity($h5p_entity, $cloned_entity_ids, 'references');

          // We need to copy h5p content files and library usage, that's not
          // done when duplicating or saving h5p entities. The content files
          // are not managed so that doesn't require any additional opeations.
          $content_dir = "public://{$h5p_path}/content/{$h5p_entity->id()}";
          if (\is_dir($content_dir)) {
            $destination = "public://{$h5p_path}/content/{$h5p_entity_clone->id()}";
            $this->fileSystem->mirror($content_dir, $destination);
          }
          try {
            $h5p->copyLibraryUsage($h5p_entity_clone->id(), $h5p_field->h5p_content_id);
          }
          catch (IntegrityConstraintViolationException $e) {
            // Do nothing if the library usage was already copied.
          }

          // The h5p field item doesn't set the default property so we can't
          // just set('opigno_h5p', $h5p_entity_clone->id()).
          $cloned_activity->set('opigno_h5p', [
            'h5p_content_id' => $h5p_entity_clone->id(),
          ]);
        }
      }

      // Update title if public_name field exists.
      // @todo: figure out how to do it elsewhere eventually.
      if ($cloned_activity->hasField('public_name')) {
        $public_name = $cloned_activity->get('public_name');
        if ($public_name->isEmpty()) {
          $cloned_activity->set('public_name', $original_activity->label());
        }
        $cloned_activity->set('name', $original_activity->label() . ' ' . $group_clone_id);
      }

      $cloned_activity->save();

      $cloned_entity_ids['opigno_activity'][$original_activity->id()] = $cloned_activity->id();

      $relationship_row['parent_id'] = $cloned_module->id();
      $relationship_row['parent_vid'] = $cloned_module->getRevisionId();
      $relationship_row['child_id'] = $cloned_activity->id();
      $relationship_row['child_vid'] = $cloned_activity->getRevisionId();
      $relationship_row['weight'] = $weight++;
      $relationship_row += $activity_data[$original_activity->id()];
      $relationship_values[] = $relationship_row;
    }

    $insert_query = $this->database->insert('opigno_module_relationship')->fields([
      'parent_id',
      'parent_vid',
      'child_id',
      'child_vid',
      'max_score',
      'group_id',
      'weight',
      'activity_status',
      'auto_update_max_score',
    ]);
    foreach ($relationship_values as $relationship_row) {
      $insert_query->values($relationship_row);
    }
    $insert_query->execute();

    return $cloned_module;
  }

}
