<?php

declare(strict_types=1);

namespace Drupal\opigno_learning_path\Plugin\GroupContentCloner;

use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\group\Entity\GroupInterface;
use Drupal\group\Entity\GroupType;
use Drupal\group_clone\Plugin\GroupContentCloner\GroupContentClonerBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Filesystem\Filesystem;

/**
 * Default group content cloner plugin.
 *
 * @GroupContentCloner(
 *   id = "tft_cloner",
 *   label = "TFT content",
 * )
 */
class TftCloner extends GroupContentClonerBase {

  private const FIELD_NAME = 'field_learning_path_folder';

  /**
   * Constructs a new instance.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    array $plugin_definition,
    EntityTypeManagerInterface $entityTypeManager,
    protected readonly Connection $database
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entityTypeManager);

    $this->fileSystem = new Filesystem();
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): self {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('database')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getConfigurationForm(
    &$element,
    FormStateInterface $form_state,
    GroupType $group_type,
    ?GroupInterface $group = NULL
  ): void {
    // Don't show settings on group clone form.
    if ($group !== NULL) {
      return;
    }

    $element['enabled'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enabled'),
      '#default_value' => $this->configuration['enabled'],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getContentIdsToClone(GroupInterface $group): array {
    if (!\array_key_exists('enabled', $this->configuration) || !$this->configuration['enabled']) {
      return [];
    }

    $clonables = [];

    $tft_field = $group->get(self::FIELD_NAME);
    if (!$tft_field->isEmpty()) {
      $clonables['tft_term'] = [$tft_field->first()->target_id];
    }

    return $clonables;
  }

  /**
   * {@inheritdoc}
   */
  public function appliesTo(GroupType $group_type): bool {
    return (
      $group_type->id() === 'learning_path' ||
      $group_type->id() === 'community'
    );
  }

  /**
   * {@inheritdoc}
   */
  public function clone(string $type, string $id, string $group_clone_id, array &$cloned_entity_ids): void {

    $term = $this->entityTypeManager->getStorage('taxonomy_term')->load($id);

    // Main term must be cloned.
    if (
      !\array_key_exists('taxonomy_term', $cloned_entity_ids) ||
      !\array_key_exists($id, $cloned_entity_ids['taxonomy_term'])
    ) {
      $group_clone = $this->entityTypeManager->getStorage('group')->load($group_clone_id);
      $this->cloneEntity($term, $cloned_entity_ids, 'all', [
        'name' => $group_clone->label(),
      ]);
    }

    $this->cloneTftTree([$term], $cloned_entity_ids);
  }

  /**
   * @param TermInterface[] $terms
   *   Array of already cloned terms.
   */
  private function cloneTftTree(array $terms, array &$cloned_entity_ids): void {
    $term_storage = $this->entityTypeManager->getStorage('taxonomy_term');
    foreach ($terms as $term) {
      $child_terms = $term_storage->loadTree('tft_tree', $term->id(), 1, TRUE);
      foreach ($child_terms as $child_term) {
        // Clone the term with all referenced entities.
        $child_clone = $this->cloneEntity($child_term, $cloned_entity_ids, 'all', [
          'parent' => $cloned_entity_ids['taxonomy_term'][$term->id()],
        ]);
      }
      // Call recursively for the cloned child term IDs to clone
      // their children.
      if (\count($child_terms) !== 0) {
        $this->cloneTftTree($child_terms, $cloned_entity_ids);
      }
    }
  }

}
