<?php

declare(strict_types=1);

namespace Drupal\opigno_learning_path;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\opigno_group_manager\Entity\OpignoGroupManagedContent;
use Drupal\opigno_learning_path\Entity\Bundle\LearningPath;
use Drupal\opigno_learning_path\Entity\LPStatus;
use Drupal\opigno_learning_path\Exception\TrainingException;
use Drupal\opigno_module\Entity\OpignoActivity;
use Drupal\opigno_module\Entity\OpignoAnswer;
use Drupal\opigno_module\Entity\OpignoModule;
use Drupal\opigno_module\Entity\UserModuleStatus;

/**
 * The Training manager.
 */
class TrainingManager {

  /**
   * The constructor.
   */
  public function __construct(
    protected readonly EntityTypeManagerInterface $entityTypeManager,
    protected readonly Connection $db,
    protected readonly TimeInterface $time,
    protected readonly LoggerChannelInterface $logger
    ) {}

  /**
   * Code saver.
   */
  private function getStorage(string $entity_type_id) {
    return $this->entityTypeManager->getStorage($entity_type_id);
  }

  /**
   * Gets a learning path status entity given a learning path and user.
   *
   * @param \Drupal\opigno_learning_path\Entity\Bundle\LearningPath $training
   *   The training.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The user account.
   * @param mixed[] $conditions
   *   Array of query conditions.
   *
   * @return \Drupal\opigno_learning_path\Entity\LPStatusInterface|null
   *   LP status entity or NULL.
   */
  public function getLearningPathStatus(LearningPath $training, AccountInterface $account, array $conditions = []): ?LPStatus {
    $lp_status_storage = $this->getStorage('user_lp_status');
    $query = $lp_status_storage->getQuery()
      ->accessCheck(FALSE)
      ->condition('gid', $training->id())
      ->condition('uid', $account->id());

    foreach ($conditions as $property => $value) {
      $query->condition($property, $value);
    }
    $query->sort('started', 'DESC');
    $query->range(0, 1);
    $result = $query->execute();

    if ($result) {
      return $lp_status_storage->load(reset($result));
    }
    return NULL;
  }

  /**
   * Sets a learning path status for a user and a group.
   *
   * @deprecated Currently used only in update hooks.
   * @see \Drupal\aft_opigno\OpignoLearningPathDataUpdater
   */
  public function setLearningPathStatus(LearningPath $training, AccountInterface $account, array $fields): LPStatusInterface {
    $lp_status = $this->getLearningPathStatus($training, $account);

    if ($lp_status !== NULL) {
      foreach ($fields as $field => $value) {
        $lp_status->set($field, $value);
      }
    }
    else {
      $lp_status = $this->getStorage('user_lp_status')->create($fields);
    }

    $lp_status->save();

    return $lp_status;
  }

  /**
   * Get a count of certificates for a user.
   */
  public function getCertificateCount(AccountInterface $account): int {
    $lp_status_storage = $this->getStorage('user_lp_status');
    $results = $lp_status_storage->getQuery()
      ->accessCheck(FALSE)
      ->condition('uid', $account->id())
      ->condition('status', 'passed')
      ->exists('gid.entity.field_certificate')
      ->condition('gid.entity.status', 1)
      ->groupBy('gid')
      ->execute();
    $statuses = $lp_status_storage->loadMultiple($results);

    $count = 0;
    foreach ($statuses as $status) {
      // Check if the user is still a member of private groups.
      $group = $status->gid->entity;
      $visibility = $group->get('field_learning_path_visibility');
      if ($visibility->value === 'private' && $group->getMember($account) === NULL) {
        continue;
      }
      $count++;
    }

    return $count;
  }

  /**
   * Get a count of passed trainings for a user.
   */
  public function getPassedTrainingsCount(AccountInterface $account): int {
    return (int) $this->getStorage('user_lp_status')->getAggregateQuery()
      ->accessCheck(FALSE)
      ->condition('uid', $account->id())
      ->condition('status', 'passed')
      ->condition('gid.entity.status', 1)
      ->groupBy('gid')
      ->count()
      ->execute();
  }

  /**
   * Helper method to get the first LP step.
   */
  public function getFirstStep(LearningPath $group): OpignoGroupManagedContent {
    $group_content_storage = $this->getStorage('opigno_group_content');

    $query = $group_content_storage->getQuery();
    $result = $query
      ->accessCheck(FALSE)
      ->condition('group_id', $group->id())
      ->notExists('parents')
      ->execute();

    if (\count($result) !== 1) {
      throw new \Exception(sprintf('Invalid content relationships data for group %d', $group->id()));
    }
    return $group_content_storage->load(\reset($result));
  }

  /**
   * Code saver.
   */
  public function loadModuleStatus(string $lp_status_id, string $module_id): ?UserModuleStatus {
    $statuses = $this->loadModuleStatusMultiple($lp_status_id, [$module_id]);
    if (\count($statuses) === 0) {
      return NULL;
    }
    return \reset($statuses);
  }

  /**
   * Code saver.
   *
   * @return \Drupal\opigno_module\Entity\UserModuleStatus[]
   *   User module status array.
   */
  public function loadModuleStatusMultiple(string $lp_status_id, array $module_ids): array {
    $user_module_status_storage = $this->getStorage('user_module_status');
    $query = $user_module_status_storage->getQuery()
      ->accessCheck(FALSE)
      ->condition('lp_status', $lp_status_id)
      ->condition('module', $module_ids, 'IN');
    $results = $query->execute();
    return $user_module_status_storage->loadMultiple($results);
  }

  /**
   * Helper method to get the next step.
   */
  public function getNextStep(OpignoGroupManagedContent $content, LPStatus $lp_status): ?OpignoGroupManagedContent {
    $group_content_storage = $this->getStorage('opigno_group_content');
    $child_contents = $group_content_storage->loadByProperties([
      'group_id' => $content->get('group_id')->target_id,
      'parents' => $content->id(),
    ]);

    if (\count($child_contents) === 0) {
      return NULL;
    }
    if (\count($child_contents) === 1) {
      return \reset($child_contents);
    }

    // Branching mode if there are more than 2 child content entitites.
    // To be implemented if needed.
    return \reset($child_contents);
  }

  /**
   * Helper method to get all training steps ordered.
   *
   * No branching support.
   *
   * @return array
   *   Steps data.
   */
  public function getOrderedSteps(LearningPath $group): array {
    $current_content = $this->getFirstStep($group);
    $group_content_storage = $this->getStorage('opigno_group_content');
    $query = $group_content_storage->getQuery();
    $results = $query
      ->accessCheck(FALSE)
      ->condition('group_id', $group->id())
      ->condition('id', $current_content->id(), '<>')
      ->execute();
    $ordered_steps = [$current_content];
    $child_contents = $group_content_storage->loadMultiple($results);
    $count = \count($child_contents);
    for ($i = 0; $i < $count; $i++) {
      foreach ($child_contents as $child_id => $child_content) {
        if ($child_content->get('parents')->first()->target_id === $current_content->id()) {
          $current_content = $child_content;
          $ordered_steps[] = $current_content;
          unset($child_contents[$child_id]);
          break;
        }
      }
    }
    // Data integrity check.
    if (\count($ordered_steps) !== $count + 1) {
      $this->logger->warning('Invalid parent-child relationships detected in learning path %id.', [
        '%id' => $group->id(),
      ]);
    }

    $module_ids = [];
    $steps_by_module = [];
    foreach ($ordered_steps as $step) {
      $module_id = $step->get('entity_id')->value;
      $module_ids[$module_id] = $module_id;
      $steps_by_module[$module_id] = $step;
    }
    unset($ordered_steps);

    $modules = $this->getStorage('opigno_module')->loadMultiple($module_ids);

    $result = [];
    foreach ($module_ids as $module_id) {
      if (!\array_key_exists($module_id, $modules)) {
        continue;
      }
      $module = $modules[$module_id];
      $randomization = $module->getRandomization();
      $result[$module_id] = [
        'label' => $module->label(),
        'randomization' => $randomization,
        'required_score' => $steps_by_module[$module_id]->getSuccessScoreMin(),
        'activities' => [],
      ];

      // If we don't have randomization, activities are not attempt-specific
      // and can be populated here.
      if ($randomization === 0) {
        $activities = $this->getModuleActivities($module);
        foreach ($activities as $activity) {
          $result[$module_id]['activities'][$activity->id()] = [
            'label' => $activity->label(),
            'activity' => $activity,
          ];
        }
      }
    }

    return $result;
  }

  /**
   * Helper method to get next module for LPStatus.
   */
  public function getNextModuleStatus(LPStatus $lp_status): ?UserModuleStatus {
    $current_module_status_field = $lp_status->get('current_module_status');
    if ($current_module_status_field->isEmpty()) {
      return NULL;
    }

    $current_module_status = $current_module_status_field->entity;

    // All this assumes we don't have the same module added to a single LP
    // more than 1 time.
    // @todo Check if that's possible and eventually prevent it or
    // switch from module ID to content ID in training routes.
    $current_content = $this->getStorage('opigno_group_content')->loadByProperties([
      'entity_id' => $current_module_status->get('module')->target_id,
      'group_id' => $lp_status->get('gid')->target_id,
      'group_content_type_id' => 'ContentTypeModule',
    ]);

    if (\count($current_content) === 0) {
      return NULL;
    }
    $content = \reset($current_content);
    $next_content = $this->getNextStep($content, $lp_status);
    if ($next_content === NULL) {
      return NULL;
    }

    $module = $this->loadModule($next_content->get('entity_id')->value);
    return $this->getModuleStatus($lp_status, $module);
  }

  /**
   * Get previous managed group content.
   */
  private function getPreviousContent(LPStatus $lp_status, string $module_id): ?OpignoGroupManagedContent {
    $group_content_storage = $this->getStorage('opigno_group_content');
    $query = $group_content_storage->getQuery()
      ->accessCheck(FALSE)
      ->condition('group_id', $lp_status->getTrainingId())
      ->condition('entity_id', $module_id)
      ->condition('group_content_type_id', 'ContentTypeModule');
    $result = $query->execute();

    if (\count($result) === 0) {
      return NULL;
    }
    $current_content = $group_content_storage->load(\reset($result));
    $parents_field = $current_content->get('parents');
    if ($parents_field->isEmpty()) {
      return NULL;
    }

    if (\count($parents_field) === 1) {
      return $parents_field->entity;
    }

    // Branching mode - currently not supported but it'll not hurt to leave
    // that code here if ever needed. In case of multiple parents return the
    // first one that was already initialized.
    foreach ($parents_field as $item) {
      $parent_content = $item->entity;
      if ($parent_content === NULL) {
        continue;
      }
      $module_status = $this->loadModuleStatus($lp_status->id(), $parent_content->get('entity_id')->value);
      if ($module_status !== NULL) {
        break;
      }
    }

    return $parent_content;
  }

  /**
   * Gets backwards navigation parameters for LPStatus.
   *
   * @return string[]
   *   Keyed array with 2 string values:
   *     - module_id (NULL if going back 1 activity),
   *     - activity_id (NULL if going back to the previous module).
   */
  public function getBackNavParameters(LPStatus $lp_status): array {
    $module_status = $lp_status->get('current_module_status')->entity;
    if ($module_status === NULL) {
      return [];
    }

    $previous_activity_id = $module_status->getPreviousActivityId();
    if ($previous_activity_id === NULL) {
      // Go back one module.
      $previous_content = $this->getPreviousContent(
        $lp_status,
        $module_status->get('module')->target_id
      );
      // Cannot go back.
      if ($previous_content === NULL) {
        return [];
      }

      // Access check.
      if (
        // Since we want to go back, assumming the last question is answered
        // here is safe even if that is not the case (the student must've had
        // the same permissions to get to this point).
        !$lp_status->getTraining()->allowArbitraryActivity(TRUE)
      ) {
        return [];
      }
      $previous_module_id = $previous_content->get('entity_id')->value;
    }
    else {
      // Access check.
      $module = $module_status->get('module')->entity;
      if (
        !$module->getBackwardsNavigation() &&
        !$lp_status->getTraining()->allowArbitraryActivity(TRUE)
      ) {
        return [];
      }
      $previous_module_id = NULL;
    }

    return [
      'module_id' => $previous_module_id,
      'activity_id' => $previous_activity_id,
    ];
  }

  /**
   * Helper method to go backwards in a learning path.
   *
   * @return mixed[]
   *   Array of redirect route options.
   */
  public function goBackwards(LPStatus $lp_status, $back_parameters): ?array {
    if (\count($back_parameters) === 0) {
      return NULL;
    }
    $current_module_id = $lp_status->get('current_module_status')->entity->get('module')->target_id;
    if (
      $back_parameters['module_id'] !== NULL &&
      $back_parameters['module_id'] !== $current_module_id
    ) {
      $module = $this->loadModule($back_parameters['module_id']);
      $module_status = $this->getModuleStatus($lp_status, $module);
      $module_status->set('current_activity', $module_status->getLastActivityId());
      $lp_status->set('current_module_status', $module_status);
      $lp_status->save();
    }
    elseif ($back_parameters['activity_id'] !== NULL) {
      $module_status = $lp_status->get('current_module_status')->entity;
      $module_status->set('current_activity', $back_parameters['activity_id']);
    }

    // We need to save since the current activity might have changed.
    $module_status->save();

    return [
      'lp_status' => $lp_status,
      'module_status' => $module_status,
      'activity_id' => $module_status->get('current_activity')->target_id,
    ];
  }

  /**
   * Module loader method.
   *
   * Needed sometimes as we don't have a standard entity reference field
   * on OpignoGroupManagedContent entities.
   *
   * @see OpignoGroupManagedContent::baseFieldDefinitions()
   */
  public function loadModule(string $id) {
    return $this->getStorage('opigno_module')->load($id);
  }

  /**
   * Perform all operations required to start a training.
   *
   * @return mixed[]
   *   Array of redirect route options.
   */
  public function initializeTraining(LearningPath $group, AccountInterface $user): array {
    // Get current step first as this may fail if data is corrupt for some
    // reason, resulting in more trash in database if done later.
    $current_step = $this->getFirstStep($group);

    // Create a LP status entity.
    $lp_status = $this->getStorage('user_lp_status')->create([
      'gid' => $group->id(),
      'uid' => $user->id(),
      'last_activity_ts' => $this->time->getRequestTime(),
      'finished' => 0,
      'status' => LPStatus::STATUS_PROGRESS,
    ]);
    $lp_status->save();

    // It should always be possible to initialize the first module (no
    // possible parent module requirement).
    $user_module_status = $this->initializeModule($lp_status, $this->loadModule($current_step->get('entity_id')->value));
    $lp_status->set('current_module_status', $user_module_status);
    $lp_status->save();

    return [
      'lp_status' => $lp_status,
      'module_status' => $user_module_status,
      'activity_id' => $user_module_status->get('current_activity')->target_id,
    ];
  }

  /**
   * Checks if a module can be initialized.
   *
   * @throws \Drupal\opigno_learning_path\Exception\TrainingException
   */
  private function canTakeModule(LPStatus $lp_status, OpignoModule $module): void {
    $previous_content = $this->getPreviousContent($lp_status, $module->id());
    // No previous content - we can always initialize the first module.
    if ($previous_content === NULL) {
      return;
    }

    $mandatory = $previous_content->isMandatory();
    $min_score = $previous_content->getSuccessScoreMin();
    if (!$previous_content->isMandatory() && $min_score === 0) {
      return;
    }

    $previous_module = $this->loadModule($previous_content->get('entity_id')->value);

    $module_status = $this->loadModuleStatus($lp_status->id(), $previous_module->id());
    if ($module_status === NULL) {
      throw new TrainingException($lp_status, $previous_module, TrainingException::REQUIRED_NOT_STARTED);
    }
    if ($mandatory && !$module_status->isFinished()) {
      throw new TrainingException($lp_status, $previous_module, TrainingException::REQUIRED_NOT_FINISHED);
    }

    // @todo Technically we should also check if the user actually can
    // complete the module once all remaining questions are evaluated.
    // For this not to cause performance issues we'd need to calculate
    // current score and add max possible score for each unevaluated
    // activity. That'd need to be implemented in self::updateModuleStatus()
    // and saved to some new module status field.
    if (!$module_status->isEvaluated()) {
      $group = $lp_status->getTraining();
      $proceed_on_unevaluated = TRUE;
      if ($group->hasField('field_proceed_on_unevaluated')) {
        $proceed_field = $group->get('field_proceed_on_unevaluated');
        if (!$proceed_field->isEmpty()) {
          $proceed_on_unevaluated = (bool) $proceed_field->value;
        }
      }

      if ($proceed_on_unevaluated) {
        return;
      }
    }

    if ($module_status->getScore() < $min_score) {
      throw new TrainingException($lp_status, $previous_module, TrainingException::INSUFFICIENT_SCORE);
    }
  }

  /**
   * Initialize module.
   *
   * Creates a new module attempt linked to a LPStatus.
   */
  public function initializeModule(LPStatus $lp_status, OpignoModule $module, ?OpignoActivity $activity = NULL): UserModuleStatus {

    // Check if requirements are met (previous module required score,
    // field_proceed_on_evaluated).
    $this->canTakeModule($lp_status, $module);

    $module_status = $this->loadModuleStatus($lp_status->id(), $module->id());
    // If there already was an attempt, restart.
    if ($module_status !== NULL) {
      // @todo Save module attempt data for analytical purposes or push to
      // a LRS when needed before deletion.
      $answers = $this->getStorage('opigno_answer')->loadByProperties([
        'user_module_status' => $module_status->id(),
      ]);
      foreach ($answers as $answer) {
        $answer->delete();
      }
      $module_status->delete();

      $this->resetModule($lp_status, $module);
    }

    // Create new module status.
    $module_status = $this->getStorage('user_module_status')->create([
      'lp_status' => $lp_status,
      'module' => $module,
      'finished' => 0,
    ]);

    $this->setModuleStatusActivities($module_status);

    // Max score calculation happens only once, when a user module status
    // is created, that's why any modifications to learning path and its
    // child entities when any user status exists should not be allowed.
    // @todo Move max score calculation here or to some module handler service.
    $max_score = $module_status->calculateMaxScore();
    $module_status->setMaxScore($max_score);

    return $module_status;
  }

  /**
   * Sets activities of a module status.
   */
  public function setModuleStatusActivities(UserModuleStatus $module_status, ?string $current_activity_id = NULL) {
    $module = $module_status->get('module')->entity;
    if ($module === NULL) {
      throw new \Exception(sprintf('User module status references invalid module %d.', $module_status->get('module')->target_id));
    }
    $activities = $this->getModuleActivities($module, FALSE);
    $activity_ids = [];
    foreach ($activities as $activity) {
      $activity_ids[] = $activity->id;
    }

    if (\count($activity_ids) === 0) {
      throw new \Exception(sprintf('Module "%s" (%d) with no activities attempted.', $module->label(), $module->id()));
    }

    $randomization = $module->getRandomization();
    if ($randomization > 0) {
      shuffle($activity_ids);
      if ($randomization === 2) {
        $random_activities = $module->get('random_activities')->value;
        $activity_ids = \array_slice($activity_ids, 0, $random_activities);
      }
    }

    $module_status->setActivityIds($activity_ids);
    $module_status->set('total_questions', \count($activity_ids));

    if ($current_activity_id === NULL) {
      $current_activity_id = reset($activity_ids);
    }
    $module_status->set('current_activity', $current_activity_id);
  }

  /**
   * Get training route parameters from LPStatus entity.
   *
   * @return array<LPStatus, UserModuleStatus, string>
   *   Array of activity form route parameters.
   */
  public function getCurrentActivityParametersFromStatus(LPStatus $lp_status): array {
    // Get current module attempt.
    $module_status_field = $lp_status->get('current_module_status');
    if (
      $module_status_field->isEmpty() ||
      $module_status_field->entity === NULL
    ) {
      $module_status = $this->getCurrentUserModuleStatus($lp_status);
      // If we have no current user module status, it means the user got to
      // the end of the learning path with some optional modules skipped due
      // to branching scenarios.
      if ($module_status === NULL) {
        return [];
      }

      $lp_status->set('current_module_status', $module_status);
      $lp_status->save();
    }
    else {
      $module_status = $module_status_field->entity;
    }

    return [
      'lp_status' => $lp_status,
      'module_status' => $module_status,
      'activity_id' => $module_status->get('current_activity')->target_id,
    ];
  }

  /**
   * Get training route parameters.
   *
   * @return array<LPStatus, UserModuleStatus, string>
   *   Array of activity form route parameters.
   */
  public function getCurrentActivityParameters(LearningPath $group, AccountInterface $user): array {
    $lp_status = $this->getLearningPathStatus($group, $user, $group->getInitialLpStatusConditions());

    if ($lp_status === NULL) {
      return $this->initializeTraining($group, $user);
    }
    return $this->getCurrentActivityParametersFromStatus($lp_status);
  }

  /**
   * Activities loader.
   */
  public function loadActivities(array $ids): array {
    return $this->getStorage('opigno_activity')->loadMultiple($ids);
  }

  /**
   * Load answer entity for a specified module status and activity.
   */
  public function loadAnswer(UserModuleStatus $module_status, OpignoActivity $activity): ?OpignoAnswer {
    $answer_storage = $this->getStorage('opigno_answer');
    $query = $answer_storage->getQuery();
    $result = $query
      ->accessCheck(FALSE)
      ->condition('user_module_status', $module_status->id())
      ->condition('activity', $activity->id())
      ->execute();
    return \count($result) === 0 ? NULL : $answer_storage->load(\reset($result));
  }

  /**
   * Get answer entity for a specified module status and activity.
   */
  public function createAnswer(UserModuleStatus $module_status, OpignoActivity $activity): OpignoAnswer {
    return $this->getStorage('opigno_answer')->create([
      // @todo Remove some useless fields.
      // @see \Drupal\opigno_module\Entity\OpignoAnswer::baseFieldDefinitions().
      'type' => $activity->getType(),
      'user_id' => $module_status->get('lp_status')->entity->getUserId(),
      'activity' => $activity->id(),
      'module' => $module_status->get('module')->target_id,
      'user_module_status' => $module_status->id(),
      'score' => 0,
    ]);
  }

  /**
   * Get user module status for a LPStatus and module.
   */
  public function getModuleStatus(LPStatus $lp_status, OpignoModule $module, ?OpignoActivity $activity = NULL): UserModuleStatus {
    $module_status = $this->loadModuleStatus($lp_status->id(), $module->id());
    if ($module_status === NULL) {
      $module_status = $this->initializeModule($lp_status, $module, $activity);
      $module_status->save();
      return $module_status;
    }

    if ($activity !== NULL) {
      $module_status->set('current_activity', $activity);
    }

    return $module_status;
  }

  /**
   * Determine current module status for a learning path status.
   *
   * Allows to find out which module a user is currently taking
   * on a given learning path.
   */
  public function getCurrentUserModuleStatus(LPStatus $lp_status): ?UserModuleStatus {
    $step = $this->getFirstStep($lp_status->get('gid')->entity);
    $module_status_storage = $this->getStorage('user_module_status');
    $module_status = NULL;

    for (;;) {
      $query = $module_status_storage->getQuery();
      $results = $query
        ->accessCheck(FALSE)
        ->condition('lp_status', $lp_status->id())
        ->condition('module', $step->get('entity_id')->value)
        ->execute();

      if (\count($results) !== 0) {
        $module_status = $module_status_storage->load(\reset($results));
        if (!$module_status->isFinished()) {
          return $module_status;
        }
      }
      else {
        return $this->initializeModule($lp_status, $this->getStorage('opigno_module')->load($step->entity_id->value));
      }

      $step = $this->getNextStep($step, $lp_status);
      if ($step === NULL) {
        break;
      }
    }

    // Revisiting a finished training case - set last module status as current
    // if not already done and return it.
    if ($lp_status->isFinished() && $module_status !== NULL) {
      $lp_status->set('current_module_status', $module_status);
      $lp_status->save();
      $module_status->set('current_activity', $module_status->getLastActivityId());
      $module_status->save();
      return $module_status;
    }

    // If there was no return until now, it means this learning path attempt is
    // finished with some optional modules missed on the way (branching).
    return NULL;
  }

  /**
   * Get all module activities.
   *
   * @todo Implement a count query method and get rid of the $load_entities
   * parameter.
   *
   * @param \Drupal\opigno_module\Entity\OpignoModule $module
   *   The LMS module.
   * @param bool $load_entities
   *   Should an array of entities or just IDs be returned.
   */
  private function getModuleActivities(OpignoModule $module, bool $load_entities = TRUE) {
    // @todo Move implementation to this class.
    return $module->getModuleActivities($load_entities);
  }

  /**
   * Get max score of an answer.
   */
  public function getActivityMaxScore(OpignoModule $module, OpignoActivity $activity): int {
    $score_query = $this->db->select('opigno_module_relationship')
      ->fields('opigno_module_relationship', ['max_score'])
      ->condition('parent_id', $module->id())
      ->condition('parent_vid', $module->getRevisionId())
      ->condition('child_id', $activity->id())
      ->condition('child_vid', $activity->getRevisionId());
    $result = $score_query->execute()->fetchField();
    return (int) $result;
  }

  /**
   * Helper method to get answers by activity ID from a module status.
   */
  public function getAnswersByActivityId(UserModuleStatus $module_status, array $activities): array {
    $answers = $this->getStorage('opigno_answer')->loadByProperties([
      'user_module_status' => $module_status->id(),
    ]);
    $answers_by_activity_id = [];
    foreach ($answers as $answer) {
      $activity_id = $answer->get('activity')->target_id;
      // Data corruption case or recalculating finished attempts of modified
      // modules which also should never happen.
      if (!\array_key_exists($activity_id, $activities)) {
        $this->logger->warning('Answer (ID: %answer_id) to a activity (ID: %activity_id) that does not exist in module (ID: %module_id) found.', [
          '%answer_id' => $answer->id(),
          '%activity_id' => $activity_id,
          '%module_id' => $module_status->get('module')->target_id,
        ]);
        continue;
      }
      $answers_by_activity_id[$activity_id] = $answer;
    }
    return $answers_by_activity_id;
  }

  /**
   * Update user module status data.
   *
   * @return bool
   *   Did the user module status change?
   */
  public function updateModuleStatus(UserModuleStatus $module_status): bool {
    $activity_ids = $module_status->getActivityIds();
    $activities = $this->getStorage('opigno_activity')->loadMultiple($activity_ids);
    $activities_count = \count($activities);
    if ($activities_count === 0) {
      throw new \Exception(sprintf('Trying to calculate module status %d for a module with no activities.', $module_status->id()));
    }

    $answers_by_activity_id = $this->getAnswersByActivityId($module_status, $activities);

    $max_score = $module_status->getMaxScore();
    $all_answered = TRUE;
    $all_evaluated = TRUE;
    $activities_count = 0;
    $answered_count = 0;
    $evaluated_count = 0;
    $obtained_score = 0;
    foreach ($activities as $activity) {
      $activities_count++;
      if (!\array_key_exists($activity->id(), $answers_by_activity_id)) {
        $all_answered = FALSE;
        $all_evaluated = FALSE;
        continue;
      }

      $answer = $answers_by_activity_id[$activity->id()];
      $answered_count++;
      if (
        $activity->hasField('opigno_evaluation_method') &&
        $activity->get('opigno_evaluation_method')->value &&
        !$answer->isEvaluated()
      ) {
        $all_evaluated = FALSE;
        continue;
      }

      $evaluated_count++;
      $obtained_score += $answer->getScore();
    }

    if ($max_score === 0) {
      $percentage = $evaluated_count / $activities_count;
    }
    else {
      $percentage = $obtained_score / $max_score;
    }
    $percentage = \round($percentage * 100);

    if (
      (int) $module_status->get('given_answers')->value === $answered_count &&
      $module_status->getScore() === $percentage &&
      $module_status->isEvaluated() === $all_evaluated
    ) {
      return FALSE;
    }

    // Set module status values.
    $module_status->set('given_answers', $answered_count);
    $module_status->setScore($percentage);

    if ($all_answered && !$module_status->isFinished()) {
      $module_status->setFinished($this->time->getRequestTime());
      $module_status->get('current_activity')->delete();
    }

    // Calculate passed percentage.
    $contents = $this->getStorage('opigno_group_content')->loadByProperties([
      'group_id' => $module_status->get('lp_status')->entity->get('gid')->target_id,
      'entity_id' => $module_status->get('module')->target_id,
      'group_content_type_id' => 'ContentTypeModule',
    ]);
    $content = \reset($contents);
    $min_score = $content->getSuccessScoreMin();
    if ($min_score === 0) {
      $pass_percentage = \round($evaluated_count / $activities_count * 100);
    }
    else {
      $pass_percentage = $percentage >= $min_score ? 100 : \round($percentage / $min_score * 100);
    }
    $module_status->set('percent', $pass_percentage);

    $module_status->setEvaluated($all_evaluated);
    $module_status->save();
    return TRUE;
  }

  /**
   * Set last activity timestamp on LP status entity.
   */
  public function setLastActivityTime(LPStatus $lp_status): void {
    $lp_status->set('last_activity_ts', $this->time->getRequestTime());
  }

  /**
   * Calculate Learning Path status data.
   *
   * @param \Drupal\opigno_learning_path\Entity\LPStatus $lp_status
   *   The LP Status entity to update.
   * @param bool $last_activity
   *   Did the user just completed the last activity?
   */
  public function updateLpStatus(LPStatus $lp_status, bool $last_activity = FALSE) {
    $contents = $this->getStorage('opigno_group_content')->loadByProperties([
      'group_id' => $lp_status->get('gid')->target_id,
      'group_content_type_id' => 'ContentTypeModule',
    ]);
    $module_ids = [];
    foreach ($contents as $content) {
      $module_id = $content->get('entity_id')->value;
      $module_ids[$module_id] = $module_id;
    }
    $modules = $this->getStorage('opigno_module')->loadMultiple($module_ids);

    $user_module_statuses = $this->getStorage('user_module_status')->loadByProperties([
      'lp_status' => $lp_status->id(),
    ]);
    $user_module_statuses_by_module_id = [];
    foreach ($user_module_statuses as $status) {
      $module_id = $status->get('module')->target_id;
      // Data corruption case or recalculating finished attempts of modified
      // learning paths which also should never happen.
      // @todo Throw an exception?
      if (!\array_key_exists($module_id, $modules)) {
        continue;
      }
      $user_module_statuses_by_module_id[$module_id] = $status;
    }

    $all_finished = TRUE;
    $all_mandatory_answered = TRUE;
    $all_mandatory_evaluated = TRUE;
    $all_mandatory_completed = TRUE;
    $total_questions_count = 0;
    $score_weighted_sum = 0;
    foreach ($contents as $content) {
      $module_id = $content->get('entity_id')->value;
      $module = $modules[$module_id];
      // @todo Use count query in a separate method.
      $module_questions_count = \count($this->getModuleActivities($module, FALSE));
      if ($module->getRandomization() === 2) {
        $random_activities = (int) $module->get('random_activities')->value;
        if ($random_activities < $module_questions_count) {
          $module_questions_count = $random_activities;
        }
      }
      $total_questions_count += $module_questions_count;

      // Module not started.
      if (!\array_key_exists($module_id, $user_module_statuses_by_module_id)) {
        $all_finished = FALSE;
        if ($content->isMandatory()) {
          $all_mandatory_answered = FALSE;
          $all_mandatory_evaluated = FALSE;
          $all_mandatory_completed = FALSE;
        }
        continue;
      }

      $module_status = $user_module_statuses_by_module_id[$module_id];
      // Module not finished.
      if (!$module_status->isFinished()) {
        $all_finished = FALSE;
        if ($content->isMandatory()) {
          $all_mandatory_answered = FALSE;
          $all_mandatory_evaluated = FALSE;
          $all_mandatory_completed = FALSE;
        }
        continue;
      }

      $module_score = $module_status->getScore();
      $score_weighted_sum += $module_score * $module_questions_count;
      if ($content->isMandatory()) {
        if (!$module_status->isEvaluated()) {
          $all_mandatory_evaluated = FALSE;
        }
        if ($module_score < $content->getSuccessScoreMin()) {
          $all_mandatory_completed = FALSE;
        }
      }
    }

    if (!$lp_status->isFinished()) {
      if (
        $all_finished ||
        // Set LP as finished if the user just answered the last activity
        // and all mandatory actifities have been finished (branching or
        // free navigation case).
        ($last_activity && $all_mandatory_answered)
      ) {
        $lp_status->setFinished($this->time->getRequestTime());
      }
    }

    // Determine the status of the learning path.
    // LP can be finished regardless of this method logic in a branching
    // scenario.
    if ($lp_status->isFinished()) {
      if ($all_mandatory_completed) {
        $status = LPStatus::STATUS_PASSED;
      }
      elseif (!$all_mandatory_evaluated) {
        $status = LPStatus::STATUS_NEEDS_EVALUATION;
      }
      else {
        $status = LPStatus::STATUS_FAILED;
      }
    }
    else {
      $status = LPStatus::STATUS_PROGRESS;
    }

    // Revisiting a training - don't update status if previously passed.
    $no_fail_mode_passed = (
      $lp_status->getStatus() === LPStatus::STATUS_PASSED &&
      $lp_status->get('gid')->entity->noFailMode()
    );
    if (!$no_fail_mode_passed) {
      $lp_status->setStatus($status);
    }

    // Calculate weighted average score of all modules.
    if ($total_questions_count === 0) {
      $lp_status->setScore(0);
    }
    else {
      $lp_status->setScore(\round($score_weighted_sum / $total_questions_count));
    }
    $lp_status->save();

    // @todo Add percent complete field to LPStatus and calculate basing on
    // total questions and answered questions count here to optimize some
    // views.
  }

  /**
   * Calculate user learning path data.
   *
   * Utility method for updating corrupt data (should be used in an post
   * update hook, hence the ID parameters).
   *
   * @return bool
   *   Was the recalculation successful?
   */
  public function updateUserLearningPathData(string $group_id, string $user_id): bool {

    $group = $this->getStorage('group')->load($group_id);
    if ($group === NULL || $group->bundle() !== 'learning_path') {
      return FALSE;
    }
    $user = $this->getStorage('user')->load($user_id);
    if ($user === NULL) {
      return FALSE;
    }

    // Recalculate all status entities for this user and group.
    $lp_statuses = $this->getStorage('user_lp_status')->loadByProperties([
      'uid' => $user->id(),
      'gid' => $group->id(),
    ]);
    if (\count($lp_statuses) === 0) {
      return FALSE;
    }

    foreach ($lp_statuses as $lp_status) {
      $module_statuses = $this->getStorage('user_module_status')->loadByProperties([
        'lp_status' => $lp_status->id(),
      ]);

      // Update module statuses.
      $status_data = [];
      foreach ($module_statuses as $module_status) {
        $this->updateModuleStatus($module_status);
        $status_data[$module_status->id()] = [
          'percent' => $module_status->get('percent')->value,
          'started' => $module_status->getCreatedTime(),
        ];
      }

      // There should be only one per LPStatus so delete all but the best or
      // the latest one.
      if (\count($status_data) > 1) {
        $preserve_id = '';
        $max_percentage = 0;
        foreach ($status_data as $id => $item) {
          if ($item['percent'] > $max_percentage) {
            $preserve_id = $id;
          }
        }
        if ($preserve_id === '') {
          $latest_ts = 0;
          foreach ($status_data as $id => $item) {
            if ($item['started'] > $latest_ts) {
              $preserve_id = $id;
            }
          }
        }

        // If $preserve_id is still set to an empty string, looks like
        // we can delete all anyway.
        foreach ($module_statuses as $id => $module_status) {
          if ($id === $preserve_id) {
            continue;
          }
          $module_status->delete();
        }
      }

      // Recalculate LP status.
      $this->updateLpStatus($lp_status);
    }
  }

  /**
   * Reset module.
   *
   * @todo Once all data is cleaned (one module status per LP status
   * and module) remove this method and references.
   */
  public function resetModule(LPStatus $lp_status, OpignoModule $module) {
    $module_status_storage = $this->getStorage('user_module_status');
    $answer_storage = $this->getStorage('opigno_answer');
    // Due to legacy reasons, we need to cleanup pre existing data to
    // avoid possible errors. Ultimately there'll be only one module status
    // per user and LPStatus.
    $module_statuses = $module_status_storage->loadByProperties([
      'lp_status' => $lp_status->id(),
      'module' => $module->id(),
    ]);
    foreach ($module_statuses as $module_status) {
      $answers = $answer_storage->loadByProperties([
        'user_module_status' => $module_status->id(),
      ]);
      foreach ($answers as $answer) {
        $answer->delete();
      }
      $module_status->delete();
    }
  }

  /**
   * Reset training.
   *
   * Development purpose.
   *
   * @todo Remove when no longer needed.
   */
  public function resetTraining($group_id, $user_id = NULL, $reset_passed = FALSE) {
    $lp_status_storage = $this->getStorage('user_lp_status');
    $module_status_storage = $this->getStorage('user_module_status');
    $answer_storage = $this->getStorage('opigno_answer');

    $lp_status_query = $lp_status_storage->getQuery();
    $lp_status_query
      ->accessCheck(FALSE)
      ->condition('gid', $group_id);
    if ($user_id !== NULL) {
      $lp_status_query->condition('uid', $user_id);
    }
    $lp_status_ids = $lp_status_query->execute();

    // If user ID is not provided and some learning path statuses are
    // completed - only clear current module and current activity values
    // if students wanted to revisit the training.
    if ($user_id === NULL && !$reset_passed) {
      foreach ($lp_status_storage->loadMultiple($lp_status_ids) as $lp_status) {
        if (
          $lp_status->getStatus() === LPStatus::STATUS_PROGRESS ||
          $lp_status->getStatus() === LPStatus::STATUS_NEEDS_EVALUATION
        ) {
          continue;
        }
        $lp_status->set('current_module_status', NULL);
        $lp_status->save();
        unset($lp_status_ids[$lp_status->id()]);
      }
    }

    $module_status_ids = [];
    if (\count($lp_status_ids) !== 0) {
      $module_status_ids = $module_status_storage->getQuery()
        ->accessCheck(FALSE)
        ->condition('lp_status', $lp_status_ids, 'IN')
        ->execute();
    }

    $answer_ids = [];
    if (\count($module_status_ids) !== 0) {
      $answer_ids = $answer_storage->getQuery()
        ->accessCheck(FALSE)
        ->condition('user_module_status', $module_status_ids, 'IN')
        ->execute();
    }

    if (\count($answer_ids) !== 0) {
      foreach ($answer_storage->loadMultiple($answer_ids) as $answer) {
        $answer->delete();
      }
    }
    if (\count($module_status_ids) !== 0) {
      foreach ($module_status_storage->loadMultiple($module_status_ids) as $module_status) {
        $module_status->delete();
      }
    }
    if (\count($lp_status_ids) !== 0) {
      foreach ($lp_status_storage->loadMultiple($lp_status_ids) as $lp_status) {
        $lp_status->delete();
      }
    }
  }

}
