<?php

namespace Drupal\opigno_module\Form;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\opigno_learning_path\Controller\TrainingControllerTrait;
use Drupal\opigno_learning_path\Entity\LPStatus;
use Drupal\opigno_learning_path\TrainingManager;
use Drupal\opigno_module\ActivityAnswerManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form controller for Answer edit forms.
 *
 * @ingroup opigno_module
 */
class OpignoAnswerForm extends ContentEntityForm {

  use TrainingControllerTrait;

  /**
   * Answer manager.
   */
  protected ActivityAnswerManager $answerManager;

  /**
   * Training manager.
   */
  protected TrainingManager $trainingManager;

  /**
   * Constructs a OpignoAnswerForm object.
   */
  public function __construct(
    EntityRepositoryInterface $entity_repository,
    EntityTypeBundleInfoInterface $entity_type_bundle_info,
    TimeInterface $time,
    ActivityAnswerManager $answerManager,
    TrainingManager $trainingManager
  ) {
    parent::__construct($entity_repository, $entity_type_bundle_info, $time);

    $this->answerManager = $answerManager;
    $this->trainingManager = $trainingManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity.repository'),
      $container->get('entity_type.bundle.info'),
      $container->get('datetime.time'),
      $container->get('plugin.manager.activity_answer'),
      $container->get('opigno_learning_path.training_manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildForm($form, $form_state);

    // Hide revision_log_message field.
    unset($form['revision_log_message']);

    $answer = $this->entity;
    $activity = $answer->getActivity();

    // Backwards navigation.
    $form['actions']['back'] = [
      '#type' => 'submit',
      '#value' => $this->t('Back'),
      '#submit' => [
        '::backwardsNavigation',
      ],
      '#access' => FALSE,
    ];

    // Conditionally allow backwards navigation.
    $back_parameters = $this->trainingManager->getBackNavParameters($answer->getUserModuleStatus()->get('lp_status')->entity);
    if (\count($back_parameters) !== 0) {
      $form['actions']['back']['#access'] = TRUE;
    }

    // Add activity specific form elements.
    $answer_activity_type = $activity->getType();
    if ($this->answerManager->hasDefinition($answer_activity_type)) {
      $answer_instance = $this->answerManager->createInstance($answer_activity_type);
      $answer_instance->answeringForm($form);
    }

    // Remove 'delete' button.
    unset($form['actions']['delete']);

    return $form;
  }

  /**
   * {@inheritdoc}
   *
   * @todo Possibly split to a second submit handler to do redirect only.
   */
  public function save(array $form, FormStateInterface $form_state) {
    $activity = $this->entity->get('activity')->entity;
    $module_status = $this->entity->getUserModuleStatus();
    $module = $module_status->getModule();
    $lp_status = $module_status->get('lp_status')->entity;

    $max_score = $this->trainingManager->getActivityMaxScore($module, $activity);

    // Determine score and evaluated property for the answer.
    $answer_activity_type = $activity->getType();

    if ($this->answerManager->hasDefinition($answer_activity_type)) {
      $answer_plugin = $this->answerManager->createInstance($answer_activity_type);
      $evaluated = $answer_plugin->evaluatedOnSave($activity);
      if ($activity->needsManualEvaluation()) {
        $score = 0;
      }
      else {
        $score = $answer_plugin->getScore($this->entity);
      }
    }
    else {
      $evaluated = TRUE;
      $score = $max_score;
    }

    $this->entity->setEvaluated($evaluated);
    $this->entity->setScore($score);
    $this->entity->save();

    // Set latest user activity time on LP status.
    $this->trainingManager->setLastActivityTime($lp_status);

    // Proceed to the next activity.
    $next_activity_id = $module_status->getNextActivityId();
    $next_module_status = NULL;
    if ($next_activity_id === NULL) {
      $this->trainingManager->updateModuleStatus($module_status);
      try {
        $next_module_status = $this->trainingManager->getNextModuleStatus($lp_status);
      }
      catch (\Exception $e) {
        $this->trainingManager->updateLpStatus($lp_status);
        $url = $this->handleError($lp_status->getTraining(), $e);
        $form_state->setRedirectUrl($url);
        return;
      }

      if ($next_module_status !== NULL) {
        $next_activity_id = $next_module_status->get('current_activity')->target_id;
        $lp_status->set('current_module_status', $next_module_status);
      }

      // We have all required properties set on LPStatus now,
      // we can recalculate and save.
      $last_activity = $next_activity_id === NULL;
      $this->trainingManager->updateLpStatus($lp_status, $last_activity);
    }
    else {
      $module_status->set('current_activity', $next_activity_id);
      $module_status->save();
      // We also need to save LP status to update last activity time.
      $lp_status->save();
      $next_module_status = $module_status;
    }

    if ($next_activity_id === NULL) {
      // @todo Redirect to results?
      if ($lp_status->getStatus() === LPStatus::STATUS_PROGRESS) {
        $this->messenger()->addWarning($this->t('Last activity completed, training still has unfinished activities.'));
      }
      elseif ($lp_status->getStatus() === LPStatus::STATUS_NEEDS_EVALUATION) {
        $this->messenger()->addStatus($this->t('Training completed, please wait for evaluation.'));
      }
      elseif ($lp_status->getStatus() === LPStatus::STATUS_PASSED) {
        $this->messenger()->addStatus($this->t('Training completed.'));
      }
      elseif ($lp_status->getStatus() === LPStatus::STATUS_FAILED) {
        $this->messenger()->addStatus($this->t('Training failed.'));
      }
      $form_state->setRedirect('entity.group.canonical', [
        'group' => $lp_status->get('gid')->target_id,
      ]);
      return;
    }

    $form_state->setRedirect('opigno_module.group.answer_form', [
      'group' => $lp_status->get('gid')->target_id,
      'opigno_module' => $next_module_status->get('module')->target_id,
      'opigno_activity' => $next_activity_id,
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function backwardsNavigation(array $form, FormStateInterface $form_state) {
    $lp_status = $this->entity->getUserModuleStatus()->get('lp_status')->entity;
    $back_parameters = $this->trainingManager->getBackNavParameters($lp_status);
    $route_parameters = $this->trainingManager->goBackwards($lp_status, $back_parameters);

    if ($route_parameters === NULL) {
      $module = $this->entity->getUserModuleStatus()->get('module')->entity;
      $activity = $this->entity->get('activity')->entity;
      $this->messenger()->addError($this->t('An error occurred, unable to go back from @module module @activity activity.', [
        '@module' => $module->label() . ' (' . $module->id() . ')',
        '@activity' => $activity->label() . ' (' . $activity->id() . ')',
      ]));
      $form_state->setRedirect('entity.group.canonical', [
        'group' => $lp_status->get('gid')->target_id,
      ]);
      return;
    }
    $form_state->setRedirect('opigno_module.group.answer_form', [
      'group' => $route_parameters['lp_status']->get('gid')->target_id,
      'opigno_module' => $route_parameters['module_status']->get('module')->target_id,
      'opigno_activity' => $route_parameters['activity_id'],
    ]);
  }

}
