<?php

namespace Drupal\opigno_team_activity\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Url;
use Drupal\group\Entity\GroupInterface;
use Drupal\opigno_team_activity\AssignmentGroups;
use Drupal\opigno_module\Entity\OpignoActivityInterface;
use Drupal\opigno_module\Entity\OpignoModuleInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Returns responses for LMS group activity routes.
 */
class GroupActivityTeamMemberManagement extends ControllerBase {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The assignment group service.
   *
   * @var \Drupal\opigno_team_activity\AssignmentGroups
   */
  protected $assignmentGroups;

  /**
   * The controller constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, AssignmentGroups $assignment_groups) {
    $this->entityTypeManager = $entity_type_manager;
    $this->assignmentGroups = $assignment_groups;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('opigno_team_activity.assignment_groups')
    );
  }

  /**
   * Title callback for controller.
   *
   * @param \Drupal\opigno_module\Entity\OpignoActivityInterface $opigno_activity
   *   The opigino activity.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   The page title.
   */
  public function title(OpignoActivityInterface $opigno_activity) {
    return $this->t('Teams for @assignment', ['@assignment' => $opigno_activity->label()]);
  }

  /**
   * Builds the response.
   */
  public function build(GroupInterface $group, OpignoModuleInterface $opigno_module, OpignoActivityInterface $opigno_activity) {
    $build = [];

    $activity_groups = $this->assignmentGroups->getActivityGroups($group, $opigno_activity);
    if (!$activity_groups) {
      $build['content'] = [
        '#type' => 'item',
        '#markup' => $this->t('No teams created yet for this activity.'),
      ];
    }
    // This essentially limits membership to a single activity per user.
    // Might need to evaluate if that is what we want or not.
    $is_activity_member = (bool) $this->assignmentGroups->getUserActivityGroups($group, $opigno_activity, $this->currentUser());
    foreach ($activity_groups as $activity_group) {
      $available = $activity_group->get('team_size')->value;
      $member_count = count($activity_group->getMembers());
      $remaining = max($available - $member_count, 0);
      $url = Url::fromRoute('entity.group.join', ['group' => $activity_group->id()], ['query' => $this->getDestinationArray()]);
      $build['content'][$activity_group->id()]['#type'] = 'container';
      $build['content'][$activity_group->id()] = [
        '#theme' => 'item_list',
        '#items' => [
          'name' => $activity_group->label(),
          'spots' => $remaining ?
            // A little ugly, but it handles the plural options.
            $this->formatPlural($member_count, '@member_count member, @remaining', '@member_count members, @remaining', [
              '@member_count' => $member_count,
              '@remaining' => $this->formatPlural($member_count, '@remaining spot left', '@remaining spots left', ['@remaining' => $remaining]),
            ]) : $this->t('@member_count members', [
              '@member_count' => $member_count,
            ]),
          'join' => [
            '#type' => 'link',
            '#title' => t('Join team'),
            '#url' => $remaining && !$is_activity_member ? $url : NULL,
          ],
        ]
      ];
    }

    return $build;
  }

}
