<?php

namespace Drupal\opigno_team_activity\Form;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\opigno_team_activity\AssignmentGroups;
use Drupal\opigno_module\Entity\OpignoActivityInterface;
use Drupal\opigno_module\Entity\OpignoModuleInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a LMS group activity form.
 */
class GroupActivityTeamGroupManagementForm extends FormBase {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The assignment group service.
   *
   * @var \Drupal\opigno_team_activity\AssignmentGroups
   */
  protected $assignmentGroups;

  /**
   * The controller constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, AssignmentGroups $assignment_groups) {
    $this->entityTypeManager = $entity_type_manager;
    $this->assignmentGroups = $assignment_groups;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('opigno_team_activity.assignment_groups')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'opigno_team_activity_team_group_management_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, OpignoModuleInterface $opigno_module = NULL, OpignoActivityInterface $opigno_activity = NULL) {
    $group_storage = $this->entityTypeManager->getStorage('group');
    $group_ids = $group_storage->getQuery()
      ->condition('team_activity', $opigno_activity->id())
      ->execute();
    $form['generate'] = [
      '#type' => 'details',
      '#title' => t('Generate teams for team activities'),
      '#open' => empty($group_ids),
      // Unfortunately, there isn't any context to guess the group from this
      // form page.
      'membership_group' => [
        '#type' => 'entity_autocomplete',
        '#title' => $this->t('Membership group'),
        '#description' => $this->t('The learning path for the user\'s training membership.'),
        '#description_display' => 'after',
        '#selection_handler' => 'default:group',
        '#target_type' => 'group',
        '#selection_settings' => [
          'target_bundles' => [
            'learning_path',
          ]
        ],
        '#required' => TRUE,
      ],
      // Same applies to class group. No context.
      'class_group' => [
        '#type' => 'entity_autocomplete',
        '#title' => $this->t('Class group'),
        '#description' => $this->t('The class for the user\'s training membership.'),
        '#description_display' => 'after',
        '#selection_handler' => 'default:group',
        '#target_type' => 'group',
        '#selection_settings' => [
          'target_bundles' => [
            'opigno_class',
          ]
        ],
      ],
      'team_size' => [
        '#type' => 'number',
        '#title' => $this->t('Team size'),
        '#description' => $this->t('Team size limit for the activity.'),
        '#description_display' => 'after',
        '#min' => 2,
        '#default_value' => 5,
      ],
      'force' => [
        '#type' => 'checkbox',
        '#title' => $this->t('Override validation if teams are already generated and force generation of more teams.')
      ],
      'actions' => [
        '#type' => 'action',
        'submit' => [
          '#type' => 'submit',
          '#button_type' => 'primary',
          '#validate' => ['::duplicatesValidate'],
          '#submit' => ['::submitForm'],
          '#value' => $this->t('Generate teams'),
        ],
      ]
    ];
    $form_state->set('activity', $opigno_activity);

    return $form;
  }

  /**
   * Validate handler for image effect.
   */
  public function duplicatesValidate($form, FormStateInterface $form_state) {
    // If the user forces creation, don't check if teams are already generated.
    if ($form_state->getValue('force')) {
      return;
    }
    $learning_path = $this->entityTypeManager->getStorage('group')->load($form_state->getValue('membership_group'));
    $activity = $form_state->get('activity');
    $class = $form_state->getValue('class_group');
    $classes = \Drupal::service('ggroup.group_hierarchy_manager')->getGroupSubgroups($learning_path->id());
    if ($class && $classes && !isset($classes[$class])) {
      $form_state->setErrorByName('class_group', $this->t('Assign teams to classes which are nested below %group.', [
        '%group' => $learning_path->label(),
        '%id' => $learning_path->id(),
      ]));
    }
    $assignment_groups = $this->assignmentGroups->getActivityGroups($learning_path, $activity);
    foreach ($assignment_groups as $assignment_group) {
      if (!$assignment_group->get('class_group')->isEmpty() && $assignment_group->get('class_group')->target_id === $class) {
        $form_state->setErrorByName('class_group', $this->t('Teams are already generated for this class.'));
      }
      elseif ($assignment_group->get('membership_group')->target_id === $learning_path->id()) {
        $form_state->setErrorByName('membership_group', $this->t('Teams are already generated for this learning path.'));
      }
    }

  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $group_storage = $this->entityTypeManager->getStorage('group');
    $activity = $form_state->get('activity');
    $class = $form_state->getValue('class_group');
    $form_state->cleanValues();
    $values = $form_state->getValues();
    /** @var \Drupal\group\Entity\GroupInterface $group */
    $group = $this->entityTypeManager->getStorage('group')->load($values['membership_group']);
    $membership_count = count($group->getMembers());
    $new_group_count = $membership_count/$values['team_size'];
    $new_group_count = (int) $new_group_count + 1;
    if ($class) {
      $class_group = $group_storage->load($class);
    }
    for ($i = 1; $i <= $new_group_count; $i++) {
      $new_group = $group_storage->create([
        'label' => implode(' | ', [
          $activity->label(),
          $group->label(),
          $i,
        ]),
        'type' => 'team_activity',
        'team_activity' => $activity,
        'membership_group' => $group->id(),
        'team_size' => $values['team_size'],
      ]);
      if (isset($class_group)) {
        $new_group->set('class_group', $class_group->id());
      }
      $new_group->save();
    }
  }

}
