<?php

namespace Drupal\opigno_webform\Plugin\Field\FieldWidget;

use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\WidgetBase;
use Drupal\Core\Form\FormState;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Validator\ConstraintViolationInterface;

/**
 * Plugin implementation of the 'entity_reference_webform_submission' widget.
 *
 * @FieldWidget(
 *   id = "entity_reference_webform_submission",
 *   label = @Translation("Webform Submission"),
 *   description = @Translation("A webform submission form embedded in the parent entity form."),
 *   field_types = {
 *     "entity_reference"
 *   }
 * )
 */
class EntityReferenceWebformSubmissionWidget extends WidgetBase {

  /**
   * Entity type manager.
   *
   * @var \Drupal\core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->entityTypeManager = $container->get('entity_type.manager');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {
    $activity = $items->getEntity()->getActivity();
    if (!$webform = $activity->get('field_webform')->entity) {
      return ['#markup' => $this->t('The webform (@webform) is broken or missing.', ['@webform' => $this->configuration['webform_id']])];
    }

    if (!$items->isEmpty()) {
      $webform_submission = $items[$delta]->entity;
    }
    else {
      $webform_submission = $this->entityTypeManager
        ->getStorage('webform_submission')
        ->create(['webform_id' => $webform->id()]);
    }

    $form_object = $this->entityTypeManager->getFormObject($webform_submission->getEntityTypeId(), 'add');
    $form_object->setEntity($webform_submission);

    $submission_form_state = new FormState();
    $submission_form_state->setFormObject($form_object);
    $submission_form = $form_object->buildForm($element, $submission_form_state);

    $element['webform'] = $submission_form['elements'];

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function errorElement(array $element, ConstraintViolationInterface $error, array $form, FormStateInterface $form_state) {
    return $element['webform'];
  }

  /**
   * {@inheritdoc}
   */
  public function massageFormValues(array $values, array $form, FormStateInterface $form_state) {
    if (!$form_state->isValidationComplete()) {
      // Only do something once form validation is complete.
      return;
    }

    /** @var \Drupal\opigno_module\Entity\OpignoAnswerInterface $answer */
    $answer = $form_state->getFormObject()->getEntity();
    $activity = $answer->getActivity();

    foreach ($values as $key => &$value) {
      if ($submission = $answer->get($this->fieldDefinition->getName())[$key]->entity) {
        $submission->setData($value['webform']);
        $submission->save();
      }
      else {
        $submission = $this->entityTypeManager
          ->getStorage('webform_submission')
          ->create([
            'data' => $value['webform'],
            'webform_id' => $activity->get('field_webform')->target_id,
          ]);
        $submission->save();
      }

      $value = [
        'entity' => $submission,
      ];
    }

    return $values;
  }

  /**
   * {@inheritdoc}
   */
  public static function isApplicable(FieldDefinitionInterface $field_definition) {
    return $field_definition->getSetting('target_type') === 'webform_submission';
  }

}
